"""Inject our constructors and define some types
"""
from typing import Dict

import yaml

from baseline.utils import get_console_logger
from odin.version import __version__

LOGGER = get_console_logger('odin', env_key='ODIN_LOG_LEVEL')

Metrics = Dict[str, float]
Path = str


class APIField:
    """Keys that we use when communicating between server and clients."""

    STATUS = 'status'
    RESPONSE = 'response'
    COMMAND = 'command'
    REQUEST = 'request'


class APIStatus:
    """Status codes used between the server and client."""

    OK = 'OK'
    ERROR = 'ERROR'
    END = 'END'


ODIN_LOGO = r"""  ____  _____ _____ _   _
 / __ \|  __ \_   _| \ | |
| |  | | |  | || | |  \| |
| |  | | |  | || | | . ` |
| |__| | |__| || |_| |\  |
 \____/|_____/_____|_| \_|
"""



"""Websocket client"""
import argparse
import asyncio
import json
import websockets
from odin import LOGGER, APIField, APIStatus
from odin.client import ODIN_URL, ODIN_PORT


async def request_pipeline_definitions(ws: str, pipeline: str) -> None:
    """Use async to open a connection to serve.py and get a pipeline definition."""
    async with websockets.connect(ws) as websocket:
        await websocket.send(json.dumps({APIField.COMMAND: 'SHOW', APIField.REQUEST: pipeline}))

        result = json.loads(await websocket.recv())
        if result[APIField.STATUS] == APIStatus.ERROR:
            LOGGER.error(result)
            return
        if result[APIField.STATUS] == APIStatus.OK:
            for file_name, file_contents in result[APIField.RESPONSE].items():
                LOGGER.info(file_name)
                LOGGER.info("=" * 100)
                LOGGER.info(file_contents)
                LOGGER.info("")


def main():
    """Use `asyncio` to connect to a websocket and request a pipeline definition."""
    parser = argparse.ArgumentParser(description='Websocket-based Pipeline finder')
    parser.add_argument('work', help='Job')
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument(
        '--scheme',
        choices={'wss', 'ws'},
        default='wss',
        help='Websocket connection protocol, use `wss` for remote connections and `ws` for localhost',
    )
    args = parser.parse_args()
    ws = f'{args.scheme}://{args.host}:{args.port}'
    asyncio.get_event_loop().run_until_complete(request_pipeline_definitions(ws, args.work))


if __name__ == '__main__':
    main()



"""HTTP client to get job data."""

import json
import asyncio
import argparse
import websockets
from odin import LOGGER, APIField, APIStatus
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME, HttpClient


async def request_data(url: str, resource: str) -> None:
    """Get k8s data for some resource.

    :param url: The location of the server
    :param resource: The name of the resource you are asking about.
    :param namespace: The namespace of the resource you are asking about.
    """
    async with websockets.connect(url) as websocket:
        await websocket.send(json.dumps({APIField.COMMAND: 'DATA', APIField.REQUEST: {'resource': resource}}))
        resp = json.loads(await websocket.recv())
        if resp[APIField.STATUS] == APIStatus.ERROR:
            LOGGER.error(resp)
            return
        if resp[APIField.STATUS] == APIStatus.OK:
            print(json.dumps(resp[APIField.RESPONSE]))


def request_data_http(url: str, resource: str) -> None:
    """Get data for a resource over HTTP

    :param url: The base URL
    :param resource: The resource ID
    """
    results = HttpClient(url).request_data(resource)
    print(json.dumps(results))


def main():
    """Websocket client for pinging odin."""
    parser = argparse.ArgumentParser(description='Websocket-based health check')
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument(
        '--scheme',
        choices={'wss', 'ws', 'http', 'https'},
        default=ODIN_SCHEME,
        help='Connection protocol, use `http` for REST, use `wss` for remote connections and `ws` for localhost',
    )
    parser.add_argument('resource', help="The name of the resource to describe")
    args = parser.parse_args()
    url = f'{args.scheme}://{args.host}:{args.port}'

    if args.scheme.startswith('ws'):
        asyncio.get_event_loop().run_until_complete(request_data(url, args.resource))
    else:
        request_data_http(url, args.resource)


if __name__ == "__main__":
    main()



"""Client to push files to the server."""

import os
import json
import argparse
from getpass import getuser
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME, HttpClient
from odin.utils.auth import get_jwt_token
from baseline.utils import color, Colors


def push_file_maybe_create_job(
    url: str, jwt_token: str, job: str, file_name: str, file_contents: str, create_job: bool
) -> None:
    """Push a file to update a remove pipeline.

    :param url: The odin-http endpoint
    :param jwt_token: The jwt token used to auth with odin
    :param job: The job definition that will be updated
    :param file_name: The name to save the file as on the remove server
    :param file_contents: The content of the file we want to upload
    """
    client = HttpClient(url, jwt_token=jwt_token)
    if create_job:
        results = client.create_job(job)
        if 'status' in results:
            status = color('Failed tor create a new job. If the job already exists, do not try to create', Colors.RED)
            print(json.dumps(results))
            print(status)
            return
    results = client.push_file(job, file_name, file_contents)
    print(json.dumps(results))


def main():
    """This should have auth around it."""
    parser = argparse.ArgumentParser(description="Upload a file to odin")
    parser.add_argument('job', help='The Job to push the file to')
    parser.add_argument('file', help='The data to upload')
    parser.add_argument('--file_name', help='The name for the file on remote, defaults to the local file name')
    parser.add_argument('--host', default=ODIN_URL, type=str, help="The odin http host")
    parser.add_argument('--port', default=ODIN_PORT, help="The odin http port")
    parser.add_argument('--token', help="File where JWT token can reside", default=os.path.expanduser("~/.odin.token"))
    parser.add_argument(
        '--create', '-c', action='store_true', help="If this is given, we will attempt to create a job with this name"
    )
    parser.add_argument('--username', '-u', help="Username", default=getuser())
    parser.add_argument('--password', '-p', help="Password")
    parser.add_argument(
        '--scheme',
        choices={'https', 'http'},
        default=ODIN_SCHEME,
        help="Use https for remote connections and http for local",
    )
    args = parser.parse_args()

    url = f'{args.scheme}://{args.host}:{args.port}'
    file_name = args.file_name if args.file_name is not None else args.file
    with open(args.file) as rf:
        file_contents = rf.read()

    jwt_token = get_jwt_token(url, args.token, args.username, args.password)
    try:
        push_file_maybe_create_job(url, jwt_token, args.job, file_name, file_contents, args.create)
    except ValueError:
        # Try deleting the token file and start again
        if os.path.exists(args.token):
            os.remove(args.token)
            jwt_token = get_jwt_token(url, args.token, args.username, args.password)
            push_file_maybe_create_job(url, jwt_token, args.job, file_name, file_contents, args.create)


if __name__ == "__main__":
    main()



"""Define a resource handler for multi-worker MPI jobs"""
from typing import List
from kubernetes import client
from odin.store import Store
from odin.k8s import Task, Status, ResourceHandler, StatusType, json_to_selector, register_resource_handler


@register_resource_handler
class MPIJobHandler(ResourceHandler):

    """Resource handler for multi-worker MPI jobs
    """

    NAME = "MPIJob"
    GROUP = "kubeflow.org"
    PLURAL = "mpijobs"
    VERSION = "v1"
    SELECTOR = "mpi_job_name"
    ROLE_SELECTOR = "mpi_role_type"
    ROLE_TYPE = "launcher"

    @property
    def kind(self) -> str:
        """Give back the k8s "kind"

        :return: kind
        :rtype: str
        """
        return MPIJobHandler.NAME

    def __init__(self, namespace):
        super().__init__(namespace)
        self.api = client.CustomObjectsApi()

    def get_api(self) -> object:
        """Get API for the resource

        :return: An API
        :rtype: object
        """
        return self.api

    def submit(self, task: Task) -> str:

        """Submit an mpi job

        :param task: The Task to launch.
        :type task: Task
        :returns: The name of the job.
        :rtype: str
        """
        secrets = self._reference_secrets(task)
        configmaps = self._generate_configmaps(task)
        pod_spec = task.to_pod_spec(task, container_name="mpi", secrets=secrets, configmaps=configmaps)
        template_metadata = client.V1ObjectMeta(name=task.name)
        template = client.V1PodTemplateSpec(metadata=template_metadata, spec=pod_spec)

        spec = {}
        spec['replicas'] = task.num_workers
        spec['restartPolicy'] = MPIJobHandler.RESTART_NEVER
        spec['template'] = template

        mpi_job_spec = {}
        mpi_job_spec['kind'] = 'MPIJob'
        mpi_job_spec['apiVersion'] = 'kubeflow.org/' + MPIJobHandler.VERSION
        mpi_job_spec['metadata'] = client.V1ObjectMeta(generate_name=task.name)
        mpi_job_spec['spec'] = spec

        mpi_job = self.api.create_namespaced_custom_object(
            MPIJobHandler.GROUP, MPIJobHandler.VERSION, self.namespace, MPIJobHandler.PLURAL, mpi_job_spec
        )
        return mpi_job['metadata']['name']

    def get_pods(self, name: str) -> List[client.models.v1_pod.V1Pod]:
        """Find all the pods contained in a tf job.

        :param name: The name of job (the jobs db resource id)
        :type name: str
        :returns: The pods in the job
        :rtype: List[client.models.v1_pod.V1Pod]
        """
        try:
            selector = json_to_selector(
                {MPIJobHandler.SELECTOR: name, MPIJobHandler.ROLE_SELECTOR: MPIJobHandler.ROLE_TYPE}
            )
            return self.core_api.list_namespaced_pod(self.namespace, label_selector=selector).items
        except client.rest.ApiException:
            return []

    def status(self, name: str, store: Store) -> Status:
        """mpi-operator seems to have no way to support querying status, instead, find out its pods' status

        :param name: The job name
        :type name: str
        :param store: The job store
        :type store: Store
        :return: A status
        :rtype: Status
        """

        resource_id = store.get(name)[Store.RESOURCE_ID]
        statuses = [p.status for p in self.get_pods(resource_id)]
        if len(statuses) != 1:
            return Status(StatusType.RUNNING, "Unknown status")
            # raise Exception("No such job found: {}".format(name))

        status = statuses[0]
        return status

    def kill(self, name: str, store: Store) -> None:
        """Kill an MPIJob

        :param name: A name
        :type name: str
        :param store: The job store
        :type store: Store
        """
        delete_options = client.V1DeleteOptions(api_version=MPIJobHandler.VERSION, propagation_policy="Background")
        resource_id = store.get(name)[Store.RESOURCE_ID]
        self.api.delete_namespaced_custom_object(
            MPIJobHandler.GROUP,
            MPIJobHandler.VERSION,
            self.namespace,
            MPIJobHandler.PLURAL,
            resource_id,
            body=delete_options,
        )



"""Auth utils"""
import os
import requests
from prompt_toolkit import prompt
from baseline.utils import str_file
from odin import LOGGER


@str_file(token_file='r')
def _read_jwt_token_from_file(token_file):
    return token_file.read()


def _authenticate(url, username, passwd):
    response = None
    url = f'{url}/v1/auth'
    try:
        response = requests.post(url, data={'username': username, 'password': passwd})
        results = response.json()
        return results['message']
    except Exception as ex:
        try:

            response = requests.post(url, json={'username': username, 'password': passwd})
            results = response.json()
            return results['message']
        except Exception as ex:
            LOGGER.error(url)
            if response:
                LOGGER.error(response.status_code)

            raise ex


@str_file(token_file='w')
def _write_jwt_file(token_file, token):
    token_file.write(token)


def get_jwt_token(url: str, token_file: str = None, username: str = None, passwd: str = None) -> str:
    """Get a JWT token to send to the server

    1. If user and passwd are given, authenticate, and save to `token_file` if not None
    2. If user and passwd are not both given but there is a `token_file`, try and use that (could file)
    3. If no user or passwd and no `token_file`, prompt for user/passwd and authenticate

    :param url: The server to authenticate to
    :param token_file: An optional path to a file containing a JWT token
    :param username: An optional username
    :param passwd: An optional passwd
    :return: A JWT token we can send
    """

    if username and passwd:
        token = _authenticate(url, username, passwd)
        _write_jwt_file(token_file, token)
        return token

    if os.path.isfile(token_file):
        token = _read_jwt_token_from_file(token_file)
        return token

    if not username:
        username = prompt('odin username: ', is_password=False)
    if not passwd:
        passwd = prompt('odin password: ', is_password=True)

    token = _authenticate(url, username, passwd)
    _write_jwt_file(token_file, token)
    return token



"""Generate odin configs from mead configs
"""
import argparse
import getpass
import inspect
import logging
import os
import re
import stat
import shutil
from copy import deepcopy
from itertools import chain
from typing import Dict, Union, Optional, List, Tuple, Any

from baseline.utils import read_config_file, str2bool, listify, import_user_module, idempotent_append
from odin import Path
from odin.utils.yaml import write_yaml


LOGGER = logging.getLogger('odin')
FILE_PERM = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH
ALWAYS = 'Always'
IF_NOT_PRESENT = 'IfNotPresent'
NEVER = 'Never'


def set_permissions(file_name: str) -> None:
    """Update permissions on a file to be rw-rw-rw-

    :param file_name: The name of the file.
    """
    if os.path.isfile(file_name):
        os.chmod(file_name, FILE_PERM)


def generate_mead_task(  # pylint: disable=too-many-locals
    template: Dict,
    task_name: str,
    config_name: str,
    mead_image: str,
    claim: str,
    datasets: Optional[str] = None,
    embeddings: Optional[str] = None,
    data_files: Optional[Dict[str, str]] = None,
    depends: Optional[Union[str, List[str]]] = None,
    gpus: int = 1,
    addons: Dict[str, str] = None,
    pull_policy: str = ALWAYS,
) -> Dict:
    """Generate a task (an element in the task list in main.yaml) that is a mead job.

    :param template: The skeleton mead task to fill in.
    :param task_name: The name of this task.
    :param config_name: The name of the mead config file for this task.
    :param mead_image: The name of the mead docker image to use.
    :param claim: The name of the pvc to use.
    :param datasets: The path to a dataset index
    :param embeddings: The path to a embeddings index
    :param data_files: A mapping of dataset overrides.
    :param depends: Tasks that this one needs to be scheduled after.
    :param gpus: The number of gpus this task wants.
    :param addons: A mapping of addon filenames to source code.
    :param pull_policy: Should k8s repull your containers.

    :returns:
        Dict, The task that runs a mead job in an odin pipeline.
    """
    template = deepcopy(template)

    template['name'] = task_name
    template['image'] = mead_image
    template['num_gpus'] = gpus
    template['mount']['claim'] = claim
    template['pull_policy'] = pull_policy

    if depends:
        template['depends'] = deepcopy(listify(depends))

    # Update the config location.
    config_idx = template['args'].index('--config') + 1
    template['args'][config_idx] = config_name

    # Update the dataset location.
    if datasets:
        template['args'].append('--datasets')
        template['args'].append(datasets)
    if embeddings:
        template['args'].append('--embeddings')
        template['args'].append(embeddings)

    if addons:
        template['args'].append("--modules")
        for addon in addons:
            template['args'].append(f"${{WORK_PATH}}/{addon}")

    data_files = data_files if data_files else {}
    for ds, df in data_files.items():
        template['args'].append(f'--mod_{ds}_file')
        template['args'].append(df)

    return template


def generate_hpctl_task(
    template: Dict,
    image: str,
    claim: str,
    name: str,
    config_file: str,
    models: List[str],
    addons: Optional[List[str]] = None,
    seed: Optional[int] = None,
    pull_policy: str = ALWAYS,
) -> Dict:
    """Generate an hpctl task based on the template.

    :param template: The template
    :param image: The image to use
    :param claim: The pvc claim to use
    :param name: What should this task is called
    :param config_file: The location of the config file to sample from
    :param models: The models that will be trained on these samples
        This controls how many configs are sampled
    :param addons: Sampling addons that are used to sample the config
    :param seed: A seed for the RNG when sampling
    :param pull_policy: When should you pull the image
    :returns: The task definition
    """
    addons = addons if addons is not None else []
    template = deepcopy(template)

    template['name'] = name
    template['image'] = image
    template['mount']['claim'] = claim
    template['pull_policy'] = pull_policy

    # Update the config location
    template['args'][0] = re.sub(r"{{sample-config}}", config_file, template['args'][0])

    for model in models:
        template['args'].append(model)

    if seed is not None:
        template['args'].extend(['--seed', seed])

    if addons:
        template['args'].extend(chain(['--modules'], addons))

    return template


def generate_template_task(
    template: Dict,
    image: str,
    claim: str,
    template_file: str,
    task: str,
    depends: Optional[Union[str, List[str]]] = None,
    pull_policy: str = ALWAYS,
) -> Tuple[Dict, str]:
    """Create a odin-template task.

    :param template: The skeleton to fill in
    :param image: The docker container to use
    :param claim: The name of the k8s data pvc
    :param template_file: The path to the sampling directive file
    :param task: The baseline task we are templating for
    :param depends: Tasks that need to be run before this one
    :param pull_policy: Should we try to pull the image

    :returns: The template file and the output file name.
    """

    template = deepcopy(template)
    template['image'] = image
    template['mount']['claim'] = claim
    template['pull_policy'] = pull_policy

    template['name'] = f'template-{task}'

    if depends:
        template['depends'] = deepcopy(listify(depends))

    template['args'][0] = template_file

    output_idx = template['args'].index('--output') + 1
    template['args'][output_idx] = re.sub(r"{{task}}", task, template['args'][output_idx])

    template['args'][template['args'].index('--task') + 1] = task

    return template, template['args'][output_idx]


def generate_chore_task(
    template: Dict, odin_image: str, claim: str, depends: Optional[Union[str, List[str]]], pull_policy: str = ALWAYS
) -> Dict:
    """Generate a task (an element in the task list in main.yaml) that is a chore job.

    :param template: The skeleton of a chore task to fill in.
    :param odin_image: The name of the odin docker image to use.
    :param claim: The name of the pvc to use.
    :param depends: Tasks that need to run before this one.
    :param pull_policy: Should k8s repull your containers.

    :returns:
        Dict, The task that will run a single chore file in an odin pipeline.
    """
    template = deepcopy(template)

    template['image'] = odin_image
    template['mount']['claim'] = claim
    template['pull_policy'] = pull_policy

    if depends:
        template['depends'] = deepcopy(listify(depends))

    return template


def generate_chore_yaml(
    template_loc: str,
    slack: bool = True,
    slack_web_hook: Optional[str] = None,
    git_commit: bool = False,
    k8s_bump: bool = False,
    selected: bool = False,
) -> List:
    """Generate the yaml for the chore file.

    :param template_loc: The location of the various pre-defined chore files.
    :param slack: Should we have a slack chore?
    :param slack_web_hook: A custom endpoint for final slack messages.
    :param git_commit: Should we have a git commit chore?
    :param k8s_bump: Should we have a chore to bump a k8s version?
    :param selected: Should we send a message about what as selected to export?

    :returns:
        List, The chores definitions. If there are no chores it return and empty list
    """
    chores = []
    git_depends = []
    if git_commit:
        chores.extend(read_config_file(os.path.join(template_loc, 'git-chore.yml')))
        git_depends = [c['name'] for c in chores]
    if slack:
        slack_chore = read_config_file(os.path.join(template_loc, 'slack-chore.yml'))
        if slack_web_hook is not None:
            slack_chore['webhook'] = slack_web_hook
        if git_depends:
            slack_chore['depends'] = deepcopy(listify(slack_chore.get('depends', [])) + git_depends)
        chores.append(slack_chore)
    if selected:
        selected_chore = read_config_file(os.path.join(template_loc, 'selected-chore.yml'))
        if slack_web_hook is not None:
            selected_chore['webhook'] = slack_web_hook
        chores.append(selected_chore)
    return chores


def generate_eval_task(  # pylint: disable=too-many-locals
    template: Dict,
    task_name: str,
    image: str,
    claim: str,
    eval_task: str,
    eval_dataset: str,
    config: Dict,
    depends: Union[str, List[str]],
    addons: Optional[List[str]] = None,
    pull_policy: str = ALWAYS,
    **kwargs,
) -> Dict:
    """Generate the task yaml for a mead eval task.

    :param template: The skeleton of the file to add
    :param task_name: The name of this mead-eval task
    :param image: The docker image to use for this task
    :param claim: The pvc to use for this task
    :param eval_task: The task you are evaluating
    :param eval_dataset: The dataset you are evaluating with
    :param config: The config of the model you are evaluating
    :param depends: Tasks that need to be run before this one.
    :param addons: addons this task needs to use.
    :param pull_policy: When should you try to pull a new container
    :param kwargs: extra argument mapping that are converted to cli args
      `--key value for key, value in kwargs.items()`
    :returns: The mead-eval yaml
    """
    depends = listify(depends)
    idempotent_append(eval_task, depends)
    config = deepcopy(config)
    template = deepcopy(template)
    template['name'] = task_name
    template['image'] = image
    template['mount']['claim'] = claim
    template['pull_policy'] = pull_policy
    template['depends'] = deepcopy(depends)

    model_idx = template['args'].index('--model') + 1
    template['args'][model_idx] = re.sub(r"{{eval-task}}", eval_task, template['args'][model_idx])

    # Point this at the bundle created by the mead-train you are testing.
    label_idx = template['args'].index('--odin:label') + 1
    template['args'][label_idx] = re.sub(r"{{eval-task}}", eval_task, template['args'][label_idx])

    # The dataset is the new evaluation dataset
    template['args'][template['args'].index('--dataset') + 1] = eval_dataset
    # Task and backend can be pulled from the config
    template['args'][template['args'].index('--task') + 1] = config['task']
    template['args'][template['args'].index('--backend') + 1] = config['backend']

    # Pull reader type from the config and set that in the args
    reader_params = config.get('reader', config.get('loader'))
    reader_type = reader_params.pop('type') if 'type' in reader_params else reader_params.pop('reader_type')
    template['args'][template['args'].index('--reader') + 1] = reader_type
    # Extract features from the config and convert to the cli format
    features = reader_params.pop("named_fields", {})
    if features:
        template['args'].append('--features')
        for idx, name in features.items():
            template['args'].append(f"{name}:{idx}")

    # Extract the pair suffix from the config and use it from cli (because the overrides can't handle lists)
    pair_suffix = reader_params.pop("pair_suffix", [])
    if pair_suffix:
        template['args'].append('--pair_suffix')
        for suf in pair_suffix:
            template['args'].append(suf)

    # Convert the reset of the reader params to cli args
    for flag, value in reader_params.items():
        template['args'].extend(chain([f"--reader:{flag}"], map(str, listify(value))))

    # Extract trainer type
    trainer = config['train'].pop("type", config['train'].pop("trainer_type", "default"))
    # Extract verbose options
    verbose_options = config['train'].pop('verbose', {})

    # Set the rest of the trainer options as cli args
    template['args'].extend(['--trainer', trainer])
    for flag, value in config['train'].items():
        template['args'].extend(chain([f"--trainer:{flag}"], map(str, listify(value))))

    # If verbose is a bool (like in tagger config) convert to dict
    if isinstance(verbose_options, bool):
        verbose_options = {'console': 1}
    # Convert verbose options to cli options
    for flag, value in verbose_options.items():
        template['args'].extend(chain([f"--verbose:{flag}"], map(str, listify(value))))

    if addons:
        module_idx = template['args'].index('--modules') + 1
        for addon in addons:
            template['args'].insert(module_idx, addon)

    # Set any kwargs to cli args.
    for flag, value in kwargs.items():
        template['args'].extend([f"--{flag}", value])

    return template


def generate_export_task(  # pylint: disable=too-many-locals
    template: Dict,
    odin_image: str,
    claim: str,
    models: List[str],
    task: str,
    dataset_name: str,
    depends: List[str],
    metric: str = 'acc',
    export_policy: Optional[str] = None,
    pull_policy: str = ALWAYS,
) -> Dict:
    """Generate a task (an element in the task list for the pipeline) that is a export job.

    :param template: The skeleton version of the export task.
    :param odin_image: The name of the odin docker image to use.
    :param claim: The name of the pvc to use.
    :param models: A list of models to evaluate.
    :param task: The name of the task that was used in training.
    :param dataset_name: The name of the dataset used to train on.
    :param depends: The tasks that need to be run before this one
    :param metric: The metric to use when comparing models.
    :param export_policy: How to make a decision if we should export things.
    :param pull_policy: Should k8s repull your containers.

    :returns:
        Dict, The task the represents a export job to run in the odin pipeline.
    """
    if not export_policy:
        return {}
    depends = listify(depends)
    template = deepcopy(template)
    template['image'] = odin_image
    template['mounts'][0]['claim'] = claim
    template['pull_policy'] = pull_policy

    # For each model we trained add it to `args` just after `--models`
    models_idx = template['args'].index('--models') + 1
    template['args'].pop(models_idx)
    for model in (f"${{PIPE_ID}}--{name}" for name in models):
        template['args'].insert(models_idx, model)

    template['args'][template['args'].index('--task') + 1] = task
    template['args'][template['args'].index('--type') + 1] = export_policy
    template['args'][template['args'].index('--metric') + 1] = metric

    template['depends'] = deepcopy(listify(depends))
    return template


def make_user_dir(root_path: Path, uname: str) -> Path:
    """Create the directory where we will store all the users pipelines.

    :param root_path: Where all the pipelines live
    :param uname: The user making the pipelines name
    :returns: The path to the user dir
    """
    user_dir = os.path.join(root_path, uname)
    if not os.path.exists(user_dir):
        os.makedirs(user_dir)
    return user_dir


def make_pipeline_dir(root_path: Path, uname: str, pipeline_name: str, clobber: bool = False) -> Path:
    """Create the directory for the pipeline at {root_path}/{uname}/{pipeline_name}."""
    pipeline_loc = os.path.join(root_path, uname, pipeline_name)
    if os.path.exists(pipeline_loc):
        if not clobber:
            raise FileExistsError(f"{pipeline_loc} already exists!")
        shutil.rmtree(pipeline_loc)
    os.makedirs(pipeline_loc)
    return pipeline_loc


def find_const_config_prop(key: str, configs: Dict) -> Any:
    """Find a value by key and make sure it is the same across all configs.

    :param key: The key to search for
    :param configs: The configs to check the value in
    :raises ValueError: If the configs have conflicting information
    :returns: The value found.
    """
    found = set(c[key] for c in configs)
    if len(found) > 1:
        raise ValueError(f"More than one {key} was found in the configs, {found}")
    return found.pop()


def generate_pipeline(  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    root_path: str,
    uname: str,
    pipeline_name: str,
    configs: Dict[str, Dict],
    datasets: Optional[Union[str, List[Dict[str, str]]]],
    embeddings: Optional[Union[str, List[Dict[str, str]]]],
    models: int = 1,
    gpus: int = 1,
    metric: str = 'acc',
    export_policy: Optional[str] = None,
    slack: bool = True,
    slack_web_hook: Optional[str] = None,
    git_commit: bool = False,
    mead_image: Optional[str] = None,
    odin_image: Optional[str] = None,
    claim_name: Optional[str] = None,
    pull_policy: str = ALWAYS,
    clobber: bool = False,
    addons: Dict[str, Dict[str, str]] = None,
    mead_eval_dataset: Optional[str] = None,
    template: Optional[Union[Dict, List, str]] = None,
    hpctl: bool = False,
    hpctl_addons: List[Dict[str, str]] = None,
    seed: Optional[int] = None,
    **kwargs,
) -> str:
    """Generate a pipeline.

    This function can generate an odin pipeline from a mead config and some configuration options.

    The pipeline will be named {uname}/{pipeline_name}

    If you use the hpctl flag then hpctl will be used in the pipeline to generate
    configs for each model on the fly. This records it's seed (which is also settable)
    to allow for reproducability

    If a mead_eval dataset is provided a mead-eval task is created for each training task.

    If an export_type is provided then an export task will be created that depends on all the
    previous mead jobs. All of the models are considered by the export decider.

    If any chores are requested (slack, git_commit, etc.) a chore task is added. This depends on
    either the export job (if it exists) otherwise it depends on the mead tasks.

    :param root_path: The location to write the pipeline to.
    :param uname: The username of the person generating the pipeline.
    :param pipeline_name: The name of the pipeline.
    :param configs: The mead configs for the each model, mapped form name to config.
    :param datasets: A custom datasets index definition (or location) to use.
    :param embeddings: A custom embeddings index definition (or location) to use.
    :param models: The number of models to train.
    :param gpus: The number of gpus to train each model with.
    :param metric: The metric to compare models when exporting.
    :param export_policy: The export criteria to use.
    :param slack: Should we send a slack notification?
    :param slack_web_hook: A custom endpoint for final slack messages.
    :param git_commit: Should we commit an exported model to git?
    :param mead_image: The docker image for the mead config.
    :param odin_image: The docker image for the odin config.
    :param claim_name: The name of the pvc to use.
    :param pull_policy: Should k8s repull your containers.
    :param clobber: Should you overwrite an old pipeline?
    :param addons: A mapping of addon file names to source code is contained
       in a mapping keyed by the model name.
    :param mead_eval_dataset: The dataset to use in mead-eval.
    :param template: The raven-template sampling directives or a file.
    :param hpctl: Should we run hpctl sampling at the start of the file?
    :param hpctl_addons: Sampling addons needed from hpctl, not support atm
    :param seed: A seed to set hpctl's RNG.
    :param kwargs: Absorb extra args for now.
    :returns: str, The name of the generated pipeline
    """

    pipeline_loc = make_pipeline_dir(root_path, uname, pipeline_name, clobber)

    pipeline = {}
    pipeline['name'] = pipeline_name if len(configs) == 1 else f"{pipeline_name}-auto"

    # Write out any addons that and config needs
    addons = addons if addons is not None else {}
    for _, addon in addons.items():
        for addon_file, addon_source in addon.items():
            addon_file = os.path.join(pipeline_loc, addon_file)
            with open(addon_file, 'w') as wf:
                wf.write(addon_source)
                set_permissions(addon_file)
    for config in configs.values():
        config.pop('modules', None)

    # Write out datasets
    if datasets:
        if isinstance(datasets, list):
            dataset_file = os.path.join(pipeline_loc, 'datasets.yml')
            write_yaml(datasets, dataset_file)
            set_permissions(dataset_file)
            datasets = os.path.join("${WORK_PATH}", "datasets.yml")
    # Write out embeddings
    if embeddings:
        if isinstance(embeddings, list):
            embeddings_file = os.path.join(pipeline_loc, 'embeddings.yml')
            write_yaml(embeddings, embeddings_file)
            set_permissions(embeddings_file)
            embeddings = os.path.join("${WORK_PATH}", "embeddings.yml")

    template_loc = os.path.join(root_path, 'templates')
    images, claims = get_images(template_loc, mead_image, odin_image, claim_name)

    templating_template = read_config_file(os.path.join(template_loc, 'templating-template.yml'))
    mead_template = read_config_file(os.path.join(template_loc, 'mead-task-template.yml'))
    mead_eval_template = read_config_file(os.path.join(template_loc, 'mead-eval-template.yml'))
    export_template = read_config_file(os.path.join(template_loc, 'export-template.yml'))
    hpctl_template = read_config_file(os.path.join(template_loc, 'hpctl-template.yml'))
    chore_template = read_config_file(os.path.join(template_loc, 'chore-template.yml'))

    task = find_const_config_prop('task', configs.values())
    dataset = find_const_config_prop('dataset', configs.values())

    all_tasks = []
    data_files = {}
    dep = None
    # Create the template task to generate datasets
    if template:
        if isinstance(template, str):
            template_file = template
        else:
            file_name = 'sample-template.yml'
            write_yaml(template, os.path.join(pipeline_loc, file_name))
            template_file = os.path.join("${WORK_PATH}", file_name)
        template_task, output_file = generate_template_task(
            templating_template, images['template'], claims['data'], template_file, task, pull_policy=pull_policy
        )
        all_tasks.append(template_task)
        data_files = {dataset: f"{output_file}.{dataset}" for dataset in ("train", "valid", "test")}
        dep = template_task['name']

    trained_models = []
    evals = []
    for config_name, config in configs.items():
        addon = addons[config_name]
        config_file = os.path.join(pipeline_loc, f"{config_name}.yml")
        write_yaml(config, config_file)
        set_permissions(config_file)
        config_file = os.path.join("${WORK_PATH}", f"{config_name}.yml")
        if hpctl:
            hpctl_task = generate_hpctl_task(
                hpctl_template,
                images['hpctl'],
                claims['data'],
                f"{config_name}-sample",
                config_file,
                [f"{config_name}-{i}" for i in range(models)],
                seed=seed,
                pull_policy=pull_policy,
            )
            all_tasks.append(hpctl_task)
            dep = hpctl_task['name']
            config_file = os.path.join("${TASK_PATH}", "config.yml")
        for i in range(models):
            task_name = f"{config_name}-{i}"
            train_task = generate_mead_task(
                mead_template,
                task_name,
                config_file,
                images['mead'],
                claims['data'],
                datasets=datasets,
                embeddings=embeddings,
                data_files=data_files,
                gpus=gpus,
                addons=addons[config_name],
                depends=dep,
                pull_policy=pull_policy,
            )
            all_tasks.append(train_task)
            trained_models.append(train_task['name'])
            if mead_eval_dataset:
                name = f"{task_name}-eval"
                eval_task = generate_eval_task(
                    template=mead_eval_template,
                    task_name=name,
                    image=images['odin'],
                    claim=claims['data'],
                    eval_task=task_name,
                    eval_dataset=mead_eval_dataset,
                    config=config,
                    depends=task_name,
                    addons=addons[config_name],
                    pull_policy=pull_policy,
                )
                all_tasks.append(eval_task)
                evals.append(eval_task['name'])

    prev_tasks = list(chain(trained_models, evals))
    export_task = generate_export_task(
        export_template,
        images['odin'],
        claims['data'],
        trained_models,
        task,
        dataset,
        depends=prev_tasks,
        metric=metric,
        export_policy=export_policy,
        pull_policy=pull_policy,
    )
    chore_depends = prev_tasks
    if export_task:
        all_tasks.append(export_task)
        chore_depends = 'export'
    chores = generate_chore_yaml(template_loc, slack, slack_web_hook, git_commit, selected=export_task)
    if chores:
        chore_task = generate_chore_task(
            chore_template, images['odin'], claims['data'], depends=chore_depends, pull_policy=pull_policy
        )
        all_tasks.append(chore_task)
        chore_file = os.path.join(pipeline_loc, 'chores.yml')
        write_yaml({'chores': chores}, chore_file)
        set_permissions(chore_file)

    pipeline['tasks'] = all_tasks
    main_file = os.path.join(pipeline_loc, 'main.yml')
    write_yaml(pipeline, main_file)
    set_permissions(main_file)

    return os.path.join(uname, pipeline_name)


def get_images(
    template_loc: Path, mead: Optional[str] = None, odin: Optional[str] = None, claim: Optional[str] = None,
) -> Tuple[Dict[str, str], Dict[str, str]]:
    """Get image names with a fallback to the images file."""
    defaults = read_config_file(os.path.join(template_loc, 'images.yml'))
    mead = mead if mead is not None else defaults['mead-image']
    odin = odin if odin is not None else defaults['odin-image']
    claim = claim if claim is not None else defaults['claim-name']
    images = {'mead': mead, 'odin': odin, 'template': defaults['template-image'], 'hpctl': defaults['hpctl-image']}
    claims = {'data': claim}
    return images, claims


def guess_export_loc(
    config: Dict, output_dir: Optional[str] = None, project: Optional[str] = None, name: Optional[str] = None
) -> Dict:
    """Try to fill in missing export information."""
    export = config.get('export', {})
    if output_dir is not None:
        export['output_dir'] = output_dir
    if project is not None:
        export['project'] = project
    if name is not None:
        export['name'] = name

    required = {'output_dir', 'project', 'name'}
    if all(r in export for r in required):
        return export

    if 'output_dir' not in export:
        export['output_dir'] = '/data/nest/models'

    dataset = config['dataset']
    parts = dataset.split(":")
    if len(parts) == 1 and any(r not in export for r in required):
        LOGGER.warning(
            "Cannot guess the export location of this config."
            " Please fill in the export section of the config, "
            "switch to the new dataset format (project:name:features:date), "
            "or use the --output-dir, --project, and --name cli flags."
        )
        return {}
    if 'project' not in export:
        export['project'] = parts[0]
    if 'name' not in export:
        export['name'] = parts[1]
    print(
        f"Planning to export to `{export['output_dir']}` in the"
        f" `{export['project']}` project under the `{export['name']}` name"
    )
    resp = input("Is this ok (y/n) ")
    if resp.lower() == "n":
        LOGGER.warning(
            "Guessed locations were unsatisfactory, please fill in the export "
            "section of the config or use the --output-dir, --project, and --name cli flags."
        )
        return {}
    return export


def preprocess_arguments(args: argparse.Namespace) -> Dict:  # pylint: disable=too-many-branches
    """Update the cli args be doing things like reading in the files they point to and things like that.

    :params args: The command line args.
    :returns: The command line args with values updated.
    """
    if args.pipeline_name is None:
        args.pipeline_name, _ = os.path.splitext(os.path.basename(args.configs[0]))
    configs = {os.path.splitext(os.path.basename(config))[0]: read_config_file(config) for config in args.configs}
    for config in configs.values():
        if args.task is None and 'task' not in config:
            LOGGER.warning("No task specified, defaulting to `classify`")
            config['task'] = 'classify'
    # If there is requested export type make sure the export section is filled in in the config.
    if args.export_policy is not None:
        export = guess_export_loc(list(configs.values())[0], args.output_dir, args.project, args.name)
        if export:
            for config in configs.values():
                config['export'] = export
        else:
            exit(1)
    if args.embeddings is not None:
        if os.path.exists(args.embeddings):
            args.embeddings = read_config_file(args.embeddings)
    if args.datasets is not None:
        if os.path.exists(args.datasets):
            args.datasets = read_config_file(args.datasets)
    # If train, valid, or tests files are provided use them to populate the datasets.
    if args.train_file is not None or args.valid_file is not None or args.test_file is not None:
        # If we are not overwriting entries in a dataset index and we don't have enough datasets listed
        if args.datasets is None and (args.train_file is None or args.valid_file is None):
            LOGGER.warning("Both a train file and a valid file are required.")
            exit(1)
        args.datasets = args.datasets if args.datasets is not None else [{}]
        dataset_label = find_const_config_prop('dataset', configs.values())
        index, dataset = next(((i, d) for i, d in enumerate(args.datasets) if d.get('label') == dataset_label), (0, {}))
        # This populate this if we are build from starch, will be the same otherwise
        dataset['label'] = dataset_label
        dataset['train_file'] = args.train_file if args.train_file is not None else dataset.get('train_file')
        dataset['valid_file'] = args.valid_file if args.valid_file is not None else dataset.get('valid_file')
        dataset['test_file'] = args.test_file if args.test_file is not None else dataset.get('test_file')
        args.datasets[index] = dataset
    dataset = find_const_config_prop('dataset', configs.values())
    if args.datasets is None and ":" not in dataset:
        LOGGER.warning(
            "You did not provide a custom dataset file and the dataset (%s) appears to be an old style dataset."
            " This means the server will most likely not be able to find this dataset.",
            dataset,
        )
    config = list(configs.values())[0]  # Hack
    config_gpus = config['train'].get('gpus', config['model'].get('gpus', config.get('gpus', None)))
    config_gpus = int(config_gpus) if config_gpus is not None else config_gpus
    # If they don't pass gpu via cli set it to match the config with a default of 1
    args.gpus = (config_gpus if config_gpus is not None else 1) if args.gpus is None else args.gpus
    if config_gpus is not None and args.gpus != config_gpus:
        LOGGER.warning(
            "The number of gpus requested via cli [%d] is not equal to number requested in the config file [%d]",
            args.gpus,
            config_gpus,
        )
        exit(1)
    addons = {}
    for config_name, config in configs.items():
        config['modules'] = list(set(chain(config.get('modules', []), args.modules)))
        addons[config_name] = {
            os.path.basename(mod.__file__): inspect.getsource(mod)
            for mod in (import_user_module(m) for m in config['modules'])
        }
    if args.template is not None and os.path.isfile(args.template):
        args.template = read_config_file(args.template)

    config = {
        'uname': args.user,
        'pipeline_name': args.pipeline_name,
        'configs': configs,
        'datasets': args.datasets,
        'embeddings': args.embeddings,
        'models': args.models,
        'gpus': args.gpus,
        'metric': args.metric,
        'export_policy': args.export_policy,
        'slack': args.slack,
        'slack_web_hook': args.slack_web_hook,
        'git_commit': False,
        'mead_image': args.mead_image,
        'odin_image': args.odin_image,
        'claim_name': args.claim_name,
        'pull_policy': args.pull_policy,
        'clobber': args.clobber,
        'addons': addons,
        'mead_eval_dataset': args.mead_eval_dataset,
        'template': args.template,
        'hpctl': args.hpctl,
        'seed': args.seed,
    }
    return config


def main():
    """Generate a pipeline locally."""
    parser = argparse.ArgumentParser()
    parser.add_argument("configs", help="The path to the mead config we want to turn into a pipeline.", nargs="+")
    parser.add_argument(
        "--root-path", "--root_path", default="/data/pipelines", help="The location of the pipelines repo."
    )
    parser.add_argument("--user", default=getpass.getuser(), help="The prefix of your pipeline name.")
    parser.add_argument(
        "--pipeline-name", "--pipeline_name", help="The name of your pipeline. Defaults to the name of the mead config."
    )
    parser.add_argument("--datasets", help="The location of your custom datasets index file.")
    parser.add_argument("--embeddings", help="The location of you custom embeddings file.")
    parser.add_argument("--train-file", "--train_file", help="The location of a custom train file")
    parser.add_argument("--valid-file", "--valid_file", help="The location of a custom valid file")
    parser.add_argument("--test-file", "--test_file", help="The location of a custom test file")
    parser.add_argument("--task", help="The baseline task for this model")
    parser.add_argument(
        "--models",
        type=int,
        default=1,
        help="The number of times to train the model" " (or the number of models to train when using hpctl).",
    )
    parser.add_argument("--gpus", type=int, help="The number of GPUs to give to each model training task.")
    parser.add_argument("--metric", default="acc", help="The name of the metric used to compare the results of models.")
    parser.add_argument(
        "--export-policy",
        "--export_policy",
        "--export-type",
        default=None,
        help="The type of decision to use when deciding if a model should be exported.",
    )
    parser.add_argument(
        "--slack",
        type=str2bool,
        default=True,
        help="Should we send a slack message when the pipeline has finished running?",
    )
    parser.add_argument(
        "--slack-web-hook", "--slack_web_hook", default=None, help="The endpoint for slack messages to go to"
    )
    # parser.add_argument("--git-commit", type=str2bool, default=False,
    #                     help="Should we commit the exported model with git?")
    parser.add_argument("--mead-image", "--mead_image", help="The name of the image to use for mead training.")
    parser.add_argument(
        "--odin-image", "--odin_image", help="The name of the image to use for odin exporting and odin chores."
    )
    parser.add_argument("--claim-name", "--claim_name", help="The name of the k8s pvc claim.")
    parser.add_argument(
        "--pull-policy",
        "--pull_policy",
        help="The pull policy to use for containers",
        choices={IF_NOT_PRESENT, ALWAYS, NEVER},
        default=ALWAYS,
    )
    parser.add_argument("--clobber", type=str2bool, default=True, help="Should we overwrite a previous pipeline?")
    parser.add_argument("--output-dir", "--output_dir", help="The base dir for where a model should be exported.")
    parser.add_argument("--project", help="The name of the project this model is for.")
    parser.add_argument("--name", help="The name of the model, i.e. intent, sf, etc.")
    parser.add_argument('--modules', help='modules to load', default=[], nargs='+', required=False)
    parser.add_argument('--mead-eval-dataset', "--mead_eval_dataset", help="The dataset to use for mead-eval")
    parser.add_argument('--template', help='An odin-template sample file.')
    parser.add_argument('--hpctl', help='Should hpctl be used in the pipeline to sample configs', action="store_true")
    parser.add_argument('--seed', help='A seed for controlling hpctl', type=int)
    args = parser.parse_args()

    config = preprocess_arguments(args)
    config['root_path'] = args.root_path

    name = generate_pipeline(**config)
    print(f"Pipeline created at {os.path.join(args.root_path, name)}")


if __name__ == "__main__":
    main()



"""Provide low-level primitives for scheduling Pods"""
# pylint: disable=too-many-lines

import os
import re
import json
from collections import namedtuple
from enum import Enum
from copy import deepcopy
from base64 import b64decode
from itertools import chain
from typing import Dict, List, Union, Optional, Any, AsyncIterator, Type, Tuple
import asyncio
from kubernetes import client, config
import requests_async as arequests
from baseline.utils import listify, optional_params, import_user_module
from odin import LOGGER
from odin.store import Store


ConfigMap = namedtuple('ConfigMap', 'path name sub_path')

HASH_TRAILING = '-hash'
CORE_MODULES = ['odin.handlers.job', 'odin.handlers.deployment', 'odin.handlers.service', 'odin.handlers.pod']
KF_MODULES = [
    'odin.handlers.mpijob',
    'odin.handlers.tfjob',
    'odin.handlers.pytorchjob',
]
ELASTIC_MODULES = [
    'odin.handlers.elasticjob',
]
DEFAULT_MODULES = CORE_MODULES + KF_MODULES
REGISTRY = 'registry'

Event = namedtuple("Event", "type reason source message timestamp")

Handle = Union[str, 'Job']
Volume = namedtuple('Volume', 'path name claim')
Status = namedtuple('Status', 'status_type message')
Secret = namedtuple('Secret', 'path name sub_path mode')
# Because not python3.7, These defaults are for the rightmost argument of Secret
DEFAULT_MODE = 0o644

SECRET_LOC = "/etc/odind/"
ODIN_CRED = "odin-cred"
ODIN_CRED_FILE = "odin-cred.yml"
SSH_KEY = "ssh-key"
SSH_KEY_FILE = "identity"
SSH_MODE = 0o400

ODIN_TASK_ENV = "ODIN_TASK_ID"
ODIN_CRED_ENV = "ODIN_CRED"


Secret.__new__.__defaults__ = ("", DEFAULT_MODE)


def populate_secret(secret_values: Dict) -> Secret:
    """Fill in default values of well known secrets.

    :param secret_values: The secrets asked for in the yaml.
    :returns: The secret values populated with defaults if they were missing.
    """
    if secret_values['name'] == ODIN_CRED:
        secret_values['path'] = secret_values.get('path', os.path.join(SECRET_LOC, ODIN_CRED_FILE))
        secret_values['sub_path'] = secret_values.get('sub_path', ODIN_CRED_FILE)
    if secret_values['name'] == SSH_KEY:
        secret_values['path'] = secret_values.get('path', os.path.join(SECRET_LOC, SSH_KEY_FILE))
        secret_values['sub_path'] = secret_values.get('sub_path', SSH_KEY_FILE)
        secret_values['mode'] = secret_values.get('mode', SSH_MODE)
    return Secret(**secret_values)


def populate_config_map(config_map_values: Dict) -> ConfigMap:
    """Populate a config map from a dictionary

    :param config_map_values: A dictionary of values
    :return: A config map
    """
    return ConfigMap(**config_map_values)


class Task:
    """An object that contains enough info to run in k8s
    """

    def __init__(
        self,
        name: str = None,
        image: str = None,
        command: Union[str, List[str]] = None,
        args: List[str] = None,
        mounts: Optional[List[Volume]] = None,
        secrets: Optional[List[Secret]] = None,
        config_maps: Optional[List[ConfigMap]] = None,
        num_gpus: int = None,
        pull_policy: str = 'IfNotPresent',
        node_selector: Optional[Dict[str, str]] = None,
        resource_type: str = "Pod",
        num_workers: int = 1,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
    ):
        """Constructor

        :param name: The name of this task
        :param image: The image to pull from the registry
        :param command: The command to execute from the container
        :param args: The arguments to supply the command
        :param mounts: Optional `Volume`s to mount
        :param secrets: Optional `Secret`s to mount
        :param config_maps: Optional `CongfigMap`s to mount
        :param num_gpus: The number of GPUs requested for this task
        :param pull_policy: pull policy, defaults to `IfNotPresent`
        :param node_selector: This is a list of labels that can be used to select a specific node
        :param resource_type: The resource type (defaults to "Pod")
        :param num_workers: If this is a distributed job (e.g TFJob), how many workers
        :param inputs: Data artifacts that this job consumes.
        :param outputs: The data artifacts that this job will generate.
        """
        self.name = name
        self.image = image
        self.command = command
        self.args = args
        self.mounts = mounts
        self.secrets = secrets
        self.config_maps = config_maps
        self.num_gpus = num_gpus
        self.pull_policy = pull_policy
        self.node_selector = node_selector
        self.resource_type = resource_type
        self.num_workers = num_workers
        self.inputs = inputs
        self.outputs = outputs

    @classmethod
    def from_dict(cls, dict_value: Dict) -> 'Job':
        """Create a `Job` from some dict read from JSON/YAML

        :param dict_value: a Dictionary
        :returns: A job instance based on data inside the dictionary.
        """
        mounts = dict_value.get('mount', dict_value.get('mounts'))
        mounts = [Volume(m['path'], m['name'], m['claim']) for m in listify(mounts)] if mounts is not None else None
        secrets = dict_value.get('secret', dict_value.get('secrets'))
        secrets = [populate_secret(s) for s in listify(secrets)] if secrets is not None else None
        config_maps = dict_value.get('config_map', dict_value.get('config_maps'))
        config_maps = [populate_config_map(cm) for cm in listify(config_maps)] if config_maps is not None else None

        return Task(
            dict_value['name'],
            dict_value['image'],
            dict_value['command'],
            dict_value.get('args', []),
            mounts,
            secrets,
            config_maps,
            dict_value.get('num_gpus', 0),
            dict_value.get('pull_policy', 'IfNotPresent'),
            dict_value.get('node_selector'),
            dict_value.get('resource_type', "Pod"),
            dict_value.get('num_workers', 1),
            dict_value.get('inputs'),
            dict_value.get('outputs'),
        )


class SubmitError(ValueError):
    """A custom error to raise when a Task can't be scheduled."""


class TaskManager:
    """Base interface for managing tasks
    """

    def __init__(self):
        """Base init for a scheduler
        """

    def submit(self, task: Task, **kwargs) -> str:
        """Submit something to k8s

        :param task: A `Task` to submit
        :param kwargs:
        :raises SubmitError: When the task can't be scheduled
        :return: This returns a unique `resource_id` within the cluster
        """

    def status(self, handle: Handle) -> Status:
        """Get progress

        :param handle: job id string or Job object
        """

    def kill(self, handle: Handle) -> Dict:
        """Kill a resource

        :param handle: job id string or Job object
        """

    async def wait_for(self, task: Task) -> None:
        """Wait for something to complete (async)

        :param task: The `Task` to wait for
        :param phases: A list of terminal phases
        :return: Nothing
        """

    def results(self, handle: Handle) -> Dict:
        """Get results from job store

        :param handle: A string id or Job
        :return: Job results dict
        """

    def get_resource_type(self, handle: Handle) -> str:
        """What type of resource is associated with this job

        :param handle: The string name of the job or a `Job` itself
        :return: A ResourceType
        """

    def find_resource_names(self, task: str) -> List[str]:
        """What are the resources associated with this task"""


class ResourceHandler:
    """Handler for a resource type
    """

    PHASE_SUCCEEDED = 'Succeeded'
    PHASE_RUNNING = 'Running'
    TERMINAL_PHASES = ['Terminated', PHASE_SUCCEEDED, 'Error', 'Failed', 'ErrImagePull']
    RESTART_NEVER = "Never"
    RESTART_ON_FAILURE = "OnFailure"

    def __init__(self, namespace: str):
        """Create a resource handler

        :param namespace: The given namespace
        :type namespace: str
        """
        self.core_api = client.CoreV1Api()
        self.namespace = namespace

    @property
    def kind(self) -> str:
        """Give back the kind.  Base throws

        :raise Exception
        """
        raise Exception("Not implemented!")

    def _reference_secrets(self, task: Task) -> Optional[List[Secret]]:
        """Generate secrets based on the requirements of the job.

        Eventually we can support custom secrets by having the job create
        secrets from the yaml config. Then this function will combine secrets
        on the job with these injected secrets to yield the final full list.

        :param task: The job we are running to add secrets to.
        :type task: Task
        :returns: A list of Secrets or `None`
        :rtype: Optional[List[Secret]]
        """
        secrets = task.secrets if task.secrets is not None else []
        command = listify(task.command)
        if command[0].startswith('odin'):
            try:
                # Check if the odin-cred secret exists
                _ = self.core_api.read_namespaced_secret(name=ODIN_CRED, namespace=self.namespace)
                cred_secret = Secret(os.path.join(SECRET_LOC, ODIN_CRED_FILE), ODIN_CRED, ODIN_CRED_FILE)
                # Make sure they aren't already requesting this secret
                if not any(s == cred_secret for s in secrets):
                    secrets.append(cred_secret)
            except client.rest.ApiException:
                if '--cred' not in task.args:
                    LOGGER.warning(
                        'No --cred arg found on job %s and no odin-cred secret found to populate container.', task.name
                    )
        if command[0].startswith('odin-chores'):
            try:
                # Check if the ssh-key secret exists
                _ = self.core_api.read_namespaced_secret(name=SSH_KEY, namespace=self.namespace)
                # Make the key permissions -rw-------
                ssh_secret = Secret(os.path.join(SECRET_LOC, SSH_KEY_FILE), SSH_KEY, SSH_KEY_FILE, SSH_MODE)
                # Make sure they aren't already requesting this secret
                if not any(s == ssh_secret for s in secrets):
                    secrets.append(ssh_secret)
            except client.rest.ApiException:
                pass
        return secrets if secrets else None

    def _generate_configmaps(self, task: Task) -> Optional[List[ConfigMap]]:

        """Generate configmaps based on the requirements of the job.

        Eventually we can support custom configmaps by having the job create
        configmaps from the yaml config. Then this function will combine configmaps
        on the job with these injected configmaps to yield the final full list.

        :param task: The job we are running and want to add configmaps too.
        :type task: Task
        :returns: A list of configmaps or `None`
        :rtype: Optional[List[ConfigMap]]
        """
        configmaps = task.config_maps if task.config_maps is not None else []
        command = listify(task.command)
        if command[0].startswith('odin-chores'):
            try:
                # Check that the ssh-config configmap exists
                _ = self.core_api.read_namespaced_config_map(name='ssh-config', namespace=self.namespace)
                # Inject an ssh_config that will use the ssh key we inject with a secret
                ssh_config = ConfigMap('/etc/ssh/ssh_config', 'ssh-config', 'ssh_config')
                # Inject a known hosts file so it can find our gitlab server.
                known_hosts = ConfigMap('/etc/ssh/ssh_known_hosts', 'ssh-config', 'known_hosts')
                configmaps.extend((ssh_config, known_hosts))
            except client.rest.ApiException:
                pass
        return configmaps if configmaps else None

    def get_events(self, name: str, store: Store) -> List[Event]:
        """Get the k8s events that happened to some pod.

        :param name: The pod to get events for.
        :param store: The job store.

        :returns: The event information.
        """

        try:
            task_entry = store.get(name)
            # If its a CRD, the resource_id may not match the user given name, so this updates it
            name = task_entry[Store.RESOURCE_ID]
        except KeyError:
            pass

        field_selectors = (
            f"involvedObject.name={name},involvedObject.namespace={self.namespace},involvedObject.kind={self.kind}"
        )
        events = self.core_api.list_event_for_all_namespaces(field_selector=field_selectors)
        return [
            Event(
                event.type,
                event.reason,
                ",".join([x for x in event.source.to_dict().values() if x is not None]),
                event.message,
                event.last_timestamp,
            )
            for event in events.items
        ]

    def get_api(self):
        """Give back the API handler for this thing
        :return:
        """

    def get_pods(self, name: str) -> List[client.models.v1_pod.V1Pod]:
        """Get pods associated with this resource

        :param name: The name of this resource
        :type name: str
        :return A list of pods associated with this Task
        :rtype: List[client.models.v1_pod.V1Pod]
        """

    def submit(self, task: Task) -> str:
        """Submit a resource

        :param task: Task definition
        :type task: Task
        :return: A string identifier
        :rtype: str
        """

    def status(self, name: str, store: Store) -> Status:
        """Get a status by resource

        :param name:  A resource identifier
        :type name: str
        :param store: A jobs store
        :type store: Store
        :return: Return a `Status` object
        """

    def kill(self, name: str, store: Store):
        """Kill a resource that is running in k8s

        :param name: A resource identifier
        :type name: str
        :param store: A jobs store
        :type store: Store
        :return: None
        """


class StatusType(Enum):
    """General purpose mapping of Status onto an object, irrespective of ResourceType

    """

    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCEEDED = "SUCCEEDED"
    MISSING = "MISSING"

    @staticmethod
    def from_pod_status(pod_status: client.V1PodStatus) -> 'StatusType':
        """Create a status message from `Pod` output

        :param pod_status: The pod status
        :return: An enum value
        """
        phase = pod_status.phase
        if phase not in ResourceHandler.TERMINAL_PHASES:
            return StatusType.RUNNING
        if phase == 'Succeeded':
            return StatusType.SUCCEEDED
        return StatusType.FAILED

    @staticmethod
    def from_job_status(job_status: Any) -> 'StatusType':
        """Create a status message from `Job` output

        :param job_status: The job status
        :return: An enum value
        """
        if job_status.active:
            return StatusType.RUNNING
        if job_status.failed:
            return StatusType.FAILED

        #  if job_status.succeeded:
        return StatusType.SUCCEEDED


def json_to_selector(selectors: Dict[str, str]) -> str:
    """Convert a json dict into a selector string."""
    return ', '.join(f"{k}={v}" for k, v in selectors.items())


def find_bearer_token(api: client.CoreV1Api, svc_acc: str, namespace: str = 'default') -> str:
    """Get the bearer token. Used when odin is running locally on a cluster with RBAC.

    :param api: The k8s api client
    :param svc_acc: The name of the service account
    :param namespace: The namespace of the service account

    :returns: The bearer token. Returns an empty string if not found.
    """
    try:
        sa = api.read_namespaced_service_account(svc_acc, namespace=namespace)
        secret = api.read_namespaced_secret(sa.secrets[0].name, namespace=namespace)
        token = b64decode(secret.data['token']).decode('utf-8')
        return f"bearer {token}"
    except client.rest.ApiException:
        return ""


def get_custom_object_name(
    api: client.CustomObjectsApi, name: str, group: str, version: str, plural: str, namespace: str = 'default'
) -> str:
    """Get the name generated by kubeflow for tf/pyt jobs.

    :param api: The k8s api to interact with tf/pyt jobs.
    :type api: client.CustomObjectsAPI
    :param name: The name that odin gives the task
    :type name: str
    :param group: The group name of the people creating the custom resource (kubeflow.org)
    :param version: The version of the api to use. (defaults to v1beta2 for kubeflow objects)
    :param plural: The name used by the custom objects in the api. TFJob -> tfjobs, PytorchJob -> pytorchjob
    :param namespace: The namespace of the job.

    :returns: The name that kubeflow gives the job.
    """
    jobs = api.list_namespaced_custom_object(group, version, namespace, plural)['items']
    job = [j for j in jobs if j['metadata']['generateName'] == name]
    if job:
        return job[0]['metadata']['name']
    return None


def task_to_pod_spec(  # pylint: disable=too-many-locals
    task: Task,
    container_name: Optional[str] = None,
    secrets: Optional[List[Secret]] = None,
    configmaps: Optional[List[ConfigMap]] = None,
) -> client.V1PodSpec:
    """Convert this job into a POD spec that a k8s scheduler can run

    :param task: name for this task
    :type task: Task
    :param container_name: name for the container if None, it will be the job name
    :type container_name: Optional[str]
    :param secrets: A list of secrets to inject into the container.
    :type secrets: Optional[List[Secret]]
    :param configmaps: A list of configmaps to inject.
    :type configmaps: Optional[List[ConfigMap]]
    :returns: A PodSpec in the k8s client API
    :rtype: client.V1PodSpec
    """
    limits = {}
    if task.num_gpus is not None:
        limits['nvidia.com/gpu'] = task.num_gpus
    resources = client.V1ResourceRequirements(limits=limits)
    volume_mounts = []
    if task.mounts is not None:
        volume_mounts.extend(client.V1VolumeMount(mount_path=m.path, name=m.name) for m in task.mounts)

    if secrets is not None:
        volume_mounts.extend(client.V1VolumeMount(mount_path=s.path, name=s.name, sub_path=s.sub_path) for s in secrets)
    if configmaps is not None:
        volume_mounts.extend(
            client.V1VolumeMount(mount_path=c.path, name=c.name, sub_path=c.sub_path) for c in configmaps
        )

    container = client.V1Container(
        args=task.args,
        command=listify(task.command),
        name=container_name if container_name else task.name,
        image=task.image,
        volume_mounts=volume_mounts if volume_mounts else None,
        image_pull_policy=task.pull_policy,
        resources=resources,
        env=[
            client.V1EnvVar(ODIN_TASK_ENV, task.name),
            client.V1EnvVar(ODIN_CRED_ENV, os.path.join(SECRET_LOC, ODIN_CRED_FILE)),
        ],
    )

    volumes = []
    if task.mounts is not None:
        for mount in task.mounts:
            pvc = client.V1PersistentVolumeClaimVolumeSource(claim_name=mount.claim)
            volumes.append(client.V1Volume(name=mount.name, persistent_volume_claim=pvc))
    if secrets is not None:
        secrets = {s.name: s for s in secrets}
        volumes.extend(
            client.V1Volume(
                name=secret.name, secret=client.V1SecretVolumeSource(secret_name=secret.name, default_mode=secret.mode)
            )
            for secret in secrets.values()
        )
    if configmaps is not None:
        volumes.extend(
            client.V1Volume(name=c, config_map=client.V1ConfigMapVolumeSource(name=c))
            for c in set(c.name for c in configmaps)
        )

    selector = task.node_selector if task.node_selector else None

    regcred = client.V1LocalObjectReference(name='registry')
    pod_spec = client.V1PodSpec(
        containers=[container],
        image_pull_secrets=[regcred],
        volumes=volumes if volumes else None,
        node_selector=selector,
        restart_policy=ResourceHandler.RESTART_NEVER,
    )
    return pod_spec


class KubernetesTaskManager(TaskManager):
    """`TaskManager` implementation to use k8s to schedule
    """

    def __init__(self, store: Store, namespace: str = 'default', modules: List[str] = DEFAULT_MODULES):
        """Create a Pod scheduler

        :param store: A job store
        :param namespace: A k8s namespace, defaults to `default`
        :param modules: A list of ResourceHandler modules to load
        """
        super().__init__()
        self.namespace = namespace
        try:
            config.load_incluster_config()
        except config.config_exception.ConfigException:
            config.load_kube_config()
        self.store = store

        for module in modules:
            import_user_module(module)
        self.handlers = {k: create_resource_handler(k, namespace) for k in RESOURCE_HANDLERS.keys()}

    async def follow_logs(  # pylint: disable=missing-yield-type-doc,too-many-locals
        self,
        resource_name: str,
        namespace: str = 'default',
        container: Optional[str] = None,
        lines: Optional[int] = None,
        service_account: str = 'odin',
    ) -> AsyncIterator[str]:
        """Stream the logs from a pod in an async way.

        The logs are streamed a line at a time (separated by `\n`)

        Note:
            Because we stream logs via async code we have to directly interact with
            the k8s rest API via the requests_async library. This requires using
            camelCase when interacting with it.

        :param resource_name: The name of the thing to get logs from
        :param namespace: The namespace the pod lives in
        :param container: The container to get logs from. Only needed when there
            are multiple pods in a single container.
        :param lines: Only grab the last {lines} entries from the log
        :param service_account: The service account used on rbac clusters.
        :returns: An async generator of strings
        """
        try:
            config.load_incluster_config()
            conf = client.Configuration()
            api = client.CoreV1Api()
        except config.config_exception.ConfigException:
            config.load_kube_config()
            conf = client.Configuration()
            api = client.CoreV1Api()
            bearer = find_bearer_token(api, service_account)
            if bearer:
                conf.api_key['authorization'] = bearer

        pods = self.find_resource_names(resource_name)
        if len(pods) > 1:
            for msg in chain([f"Found {len(pods)} pods,"], pods, [f"Using pod/{pods[0]}"]):
                yield msg
        pod = pods[0] if pods else resource_name

        params = {'follow': 'true'}
        if container is not None:
            params['container'] = container
        if lines is not None:
            # Here we use camelCase because we are directly interacting with the k8s rest api
            # and it demands camelCase like the yaml manifests do.
            params['tailLines'] = lines
        resp = await arequests.get(
            f"{conf.host}/api/v1/namespaces/{namespace}/pods/{pod}/log",
            params=params,
            stream=True,
            verify=conf.ssl_ca_cert,
            headers=conf.api_key,
        )
        line = []
        async for chunk in resp.iter_content():
            if chunk == b"\n":
                yield b"".join(line).decode('utf-8')
                line = []
                continue
            line.append(chunk)
        yield b"".join(line).decode('utf-8')

    # Need to push this down into the resource handler
    def get_logs(self, name: str, container: Optional[str] = None, lines: Optional[int] = None) -> str:
        """Get a snapshot of the logs of a pod.

        Note:
            This function is implemented via the python k8s client which is synchronous
            and uses snake_case for parameters.

        :param name: The name of the resource to get logs from
        :param container: The container to get logs from. Only needed when there
            are multiple pods in a single container.
        :param lines: Only grab the last {lines} entries from the log
        :returns: The logs in a single string.
        """
        try:
            config.load_incluster_config()
        except config.config_exception.ConfigException:
            config.load_kube_config()
        prefix = ""
        api = client.CoreV1Api()

        pods = self.find_resource_names(name)
        if len(pods) > 1:
            prefix = "\n".join(chain([f"Found {len(pods)} pods,"], pods, [f"using pod/{pods[0]}"]))
        pod = pods[0] if pods else name

        # Here we fill a dict with the extra parameters for our call to `read_namespaced_pod_log`. This is
        # because to get the default behavior you need to not pass a value, passing a value of `None` isn't
        # the same thing. The other option is a large mess of ifs to make calls with different parameters.
        args = {}
        if container is not None:
            args['container'] = container
        if lines is not None:
            # Here the parameter is in snake_case because we are interacting with k8s through the python client.
            args['tail_lines'] = lines

        logs = api.read_namespaced_pod_log(pod, namespace=self.namespace, **args)

        return f"{prefix}\n{logs}" if prefix else logs

    def _find_resource(self, task: str) -> Tuple[str, str]:

        """Figure out what type of resource this is.

        If its a svc or deployment it should have a long name with `/` delimiter.
        If it is a job inside odin, we will just look up its kind

        :param task:
        :return: give back a 2 strings, the first is the resource type, the second is the id
        """

        if '/' in task:
            prefix, resource = task.split('/')
            prefix = prefix.lower()
            if prefix in ('svc', 'service'):
                return 'svc', resource
            if prefix in ('deploy', 'deployment'):
                return 'deploy', resource

        try:
            job_entry = self.store.get(task)
        except KeyError:
            # If the resource is missing from the job db we assume it some thing the user
            # figured out like the `${PyTorchJob_ID}-worker-0`. This lets us get logs
            # from arbitrary resources on the cluster
            return 'Pod', task

        return job_entry.get(Store.RESOURCE_TYPE, 'Pod'), job_entry[Store.RESOURCE_ID]

    def find_resource_names(self, task: str) -> List[str]:
        """Find possible pods that we could get logs from.

        If they request logs from a service with svc/service-name or deploy/deployment-name
        do the same thing kubectl logs does, pick one and show the logs. Differently from kubectl
        we list that possible options so users can get logs from one specifically if they want.

        If the resource they want logs from is a pytorch or tf job when we get logs from a specific pod

        :param task: The thing we want logs from
        :returns: The list of pods that it makes sense to get logs from.
        """
        resource_type, resource = self._find_resource(task)
        pods = self.handler_for(resource_type).get_pods(resource)
        return [p.metadata.name for p in pods]

    def handler_for(self, resource_type: str = "Pod") -> object:
        """Get the right API based on the `resource_type`

        :param resource_type: The resource type to get an API for
        :return: return some API
        """
        return self.handlers[resource_type]

    # def _get_pods(self):
    #    results = self.api_for(ResourceType.POD).list_namespaced_pod(namespace=self.namespace)
    #    return [item for item in results.items if self.store.is_a_child(item)]

    def _status(self, handle: Handle):  # pylint: disable=missing-return-type-doc
        """Get the status of a job.

        :param handle: A string id of Job
        :returns: The status object
        """
        resource_type = self.get_resource_type(handle)
        name = handle.name if isinstance(handle, Task) else handle
        handler = self.handler_for(resource_type)
        return handler.status(name, self.store)

    def status(self, handle: Handle) -> Status:
        """Get the status of a pod

        :param handle: A string id or Job
        :return: The k8s status
        """
        try:
            status = self._status(handle)
            if isinstance(status, Status):
                return status
            return Status(StatusType.from_pod_status(status), status.message)
        except client.rest.ApiException:
            return Status(StatusType.MISSING, "resource not found")

    def submit(self, task: Task, **kwargs) -> str:
        """Submit job as a pod

        :param task: a `Task` to submit
        :param kwargs:
        :return: k8s response
        """
        try:
            return self.handler_for(task.resource_type).submit(task)
        except client.rest.ApiException as exc:
            raise SubmitError(json.loads(exc.body)['message'])

    def kill(self, handle: Handle) -> None:  # pylint: disable=missing-return-type-doc
        """Kill a Job

        :param handle: A string id or Job
        """
        resource_type = self.get_resource_type(handle)
        name = handle.name if isinstance(handle, Task) else handle
        self.handler_for(resource_type).kill(name, self.store)

    async def wait_for(self, task: Task) -> Task:
        """Wait for pod to complete, as defined by any of these phases

        :param job: The `Job` to wait on
        :returns: The Job when it is done.
        """
        while self.status(task).status_type is StatusType.RUNNING:
            await asyncio.sleep(1)
        return task

    async def wait_until_running(self, task: Task) -> None:
        """Wait for a pod to actually start running.

        :param job: The `Job` to wait on.
        """
        while self._status(task).phase != ResourceHandler.PHASE_RUNNING:
            await asyncio.sleep(1)

    async def hash_task(self, task: Task) -> List[str]:
        """Get the hash of a task, defined as the list of hashes for each container the task uses.

        In order to tell if a task in a pipeline can be skipped we need to know that the code for that task is
        unchanged. This can be done with a hash of the container the code runs in. Docker creates a hash of the
        container (using sha256). k8s also tracks this container hash for containers that are running in a pod.

        In order to get the container hashes we need the job running on the cluster so we create a copy of the
        job, change the command to sleep rather than running code, and removing things that would stop it from
        running right away (gpus, node_selectors, etc) and execute this job. Now that the job is running we
        can query the k8s api to get the job status and extract the hashes for all the containers in the job.

        :param task: The task to hash
        :returns: A list of hashes for the container that make up the job.
        """
        task = deepcopy(task)
        # Update the job name so in cause it is still alive when the real job runs the names won't collide.
        task.name = task.name + HASH_TRAILING
        task.args = ['300']
        task.command = 'sleep'
        # Remove things that could stop it from running right away
        task.node_selector = None
        task.num_gpus = None
        task.mount = None
        # Spin up the task
        try:
            self.submit(task)
        except client.rest.ApiException as exc:
            raise SubmitError(json.loads(exc.body)['message'])
        # Make sure the task is actually running. We can't get the hashes if the pod is in the Pending state
        await self.wait_until_running(task)
        # Get the imageId of all the containers in the task (they look like this)
        # "docker-pullable://localhost:32000/blester/mongo-demo@sha256:550892bae020e5ac0b850d6a2e1bd0e8cc5c9eb5eed903dc3544f808981f7076"  # pylint: disable=line-too-long
        task_status = self._status(task)
        statuses = sorted(task_status.container_statuses, key=lambda x: x.image)
        hashes = list(
            map(
                lambda x: x.groups()[0],
                filter(lambda x: x is not None, (re.search('@sha256:(.*)$', s.image_id) for s in statuses)),
            )
        )
        self.kill(task)
        return hashes

    def results(self, handle: Handle) -> Dict:
        """Get results from job store

        :param handle: A string id or Job
        :return: task results dict
        """
        name = handle.name if isinstance(handle, Task) else handle
        return self.store.get(name)

    def get_events(self, name: str) -> List[Event]:
        """Get the k8s events that happened to some pod.

        :param name: name

        :returns: The event information.
        """
        try:
            config.load_incluster_config()
        except config.config_exception.ConfigException:
            config.load_kube_config()
        resource_type, name = self._find_resource(name)

        handler = self.handler_for(resource_type.lower())
        return handler.get_events(name, self.store)

        # try:
        #    job_entry = store.get(name)
        #    resource = job_entry[Store.RESOURCE_ID]
        # except KeyError:
        #    pass

    def get_resource_type(self, handle: Handle) -> str:
        """What type of resource is associated with this job

        If the input is a `Job`, we can just pass back its
        `resource_name` field.  If its a handle, we need to find
        it in the Job DB.

        :param handle: The string name of the job or a `Task` itself
        :return: A ResourceType
        """
        if isinstance(handle, Task):
            return handle.resource_type
        json_obj = self.store.get(handle)
        return json_obj.get(Store.RESOURCE_TYPE, "Pod")


RESOURCE_HANDLERS = {}


@optional_params
def register_resource_handler(cls: Type[ResourceHandler], aliases: List[str] = None) -> Type[ResourceHandler]:
    """Register an export policy

    :param cls: The class name
    :param aliases: `List[str]` A list of aliases other than ResourceHandler.NAME
    :raises Exception: If name is already registered.
    :return: The class
    """
    aliases = aliases if aliases is not None else []
    names = aliases + [cls.NAME, cls.NAME.lower()]

    for name in names:
        if name in RESOURCE_HANDLERS:
            raise Exception(
                'Error: attempt to re-define previously '
                f'registered ExportPolicy {name} (old: {RESOURCE_HANDLERS[name]}, new: {cls}) in registry'
            )
        RESOURCE_HANDLERS[name] = cls
    return cls


def create_resource_handler(name: str, namespace: str) -> ResourceHandler:
    """Create an export policy from the registry

    :param name: A name identifier
    :type name: str
    :param namespace: A namespace to pass
    :type namespace: str
    :return: A constructed instance
    """
    return RESOURCE_HANDLERS[name](namespace)



"""Client to create or update odin user"""
import argparse
import os
import json
import signal
import requests
from prompt_toolkit import prompt
from odin import LOGGER
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME
from odin.utils.auth import get_jwt_token


def create_user_http(url: str, jwt_token: str, username: str, password: str, firstname: str, lastname: str) -> None:
    """Create or update a user over HTTP
    :param url: the base URL
    :param jwt_token: The JWT token representing this authentication
    :param username: The user ID
    :param password: The updated password
    :param firstname: The firstname
    :param lastname: The lastname
    """
    user = {"username": username, "password": password}
    if firstname:
        user['firstname'] = firstname
    if lastname:
        user['lastname'] = lastname
    headers = {'Authorization': f'Bearer {jwt_token}'}

    try:
        response = requests.get(f'{url}/v1/users/{username}')
        if response.status_code == 401:
            raise ValueError("Invalid login")
        if response.status_code != 200:
            # No such user exists so do a POST
            response = requests.post(f'{url}/v1/users', headers=headers, json={"user": user})
            if response.status_code != 200:
                raise Exception(f"Failed to create user: {username}")
            results = response.json()
            LOGGER.info("Created new user")
            LOGGER.info(json.dumps(results))
            return

        results = response.json()
        LOGGER.info("Found existing user")
        LOGGER.info(json.dumps(results))
    except Exception as ex:
        LOGGER.error(ex)
        return

    response = requests.put(f'{url}/v1/users/{username}', json=user, headers=headers)
    results = response.json()
    LOGGER.info(json.dumps(results))


def main():
    """Create a new user or update an existing one.

    This requires a valid JWT token which you can get with `odin-auth`, or if it doesnt exist, it will prompt you
    for these
    """
    signal.signal(signal.SIGINT, lambda *args, **kwargs: exit(0))

    parser = argparse.ArgumentParser(description='Create or update an odin user')
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument('--token', help="File where JWT token can reside", default=os.path.expanduser("~/.odin.token"))
    parser.add_argument('--username', '-u', help="Create or update a username")
    parser.add_argument('--password', '-p', help="New or updated password")
    parser.add_argument('--firstname', '-f', help="First name")
    parser.add_argument('--lastname', '-l', help="Last name")
    parser.add_argument('--scheme', choices={'http', 'https'}, default=ODIN_SCHEME, help='The protocol to communicate over')
    args = parser.parse_args()

    if not args.username:
        args.username = prompt('create username: ', is_password=False)
    if not args.password:
        args.password = prompt('new password: ', is_password=True)

    url = f'{args.scheme}://{args.host}:{args.port}'
    jwt_token = get_jwt_token(url, args.token, None, None)
    try:
        create_user_http(url, jwt_token, args.username, args.password, args.firstname, args.lastname)
    except ValueError:
        # Try deleting the token file and start again
        if os.path.exists(args.token):
            os.remove(args.token)
            jwt_token = get_jwt_token(url, args.token, None, None)
            create_user_http(url, jwt_token, args.username, args.password, args.firstname, args.lastname)


if __name__ == '__main__':
    main()



"""Defines a resource handler for multi-worker TFJobs"""

from typing import List
from kubernetes import client
from odin.store import Store
from odin.k8s import (
    Task,
    Status,
    ResourceHandler,
    StatusType,
    task_to_pod_spec,
    json_to_selector,
    register_resource_handler,
)


@register_resource_handler
class TFJobHandler(ResourceHandler):
    """Resource handler for multi-worker PyTorchJobs"""

    NAME = "TFJob"
    GROUP = "kubeflow.org"
    PLURAL = "tfjobs"
    VERSION = "v1"
    ALIAS = ('tfjob', 'tensorflowjob')
    # This label selector was changed on Mar 7, 2019. We are using the old format
    # https://github.com/kubeflow/tf-operator/pull/951
    SELECTOR = "tf-job-name"
    GROUP_KEY = "group-name"

    @property
    def kind(self) -> str:
        """Give back the k8s "kind"

        :return: kind
        :rtype: str
        """
        return TFJobHandler.NAME

    def __init__(self, namespace):
        super().__init__(namespace)
        self.api = client.CustomObjectsApi()

    def get_api(self) -> object:
        """Get the API for this resource type

        :return An API object
        :rtype object
        """
        return self.api

    def get_pods(self, name: str) -> List[client.models.v1_pod.V1Pod]:
        """Find all the pods contained in a tf job.

        :param name: The name of job (the jobs db resource id)
        :type name: str
        :returns: The pods in the job
        :rtype: List[client.models.v1_pod.V1Pod]
        """
        try:
            selector = json_to_selector({TFJobHandler.SELECTOR: name, TFJobHandler.GROUP_KEY: TFJobHandler.GROUP})
            return self.core_api.list_namespaced_pod(self.namespace, label_selector=selector).items
        except client.rest.ApiException:
            return []

    def submit(self, task: Task) -> str:
        """Submit a multi-worker TF Task

        :param task: A task definition to run
        :type task: Task
        :return: A task identifier
        :rtype: str
        """
        secrets = self._reference_secrets(task)
        configmaps = self._generate_configmaps(task)
        task.num_gpus = 1

        pod_spec = task_to_pod_spec(task, container_name="tensorflow", secrets=secrets, configmaps=configmaps)
        template_metadata = client.V1ObjectMeta(name=task.name)

        template = client.V1PodTemplateSpec(metadata=template_metadata, spec=pod_spec)

        worker_replica_spec = {}
        worker_replica_spec['replicas'] = task.num_workers
        worker_replica_spec['restartPolicy'] = TFJobHandler.RESTART_NEVER
        worker_replica_spec['template'] = template

        spec = {}
        spec['tfReplicaSpecs'] = {}
        spec['tfReplicaSpecs']['Worker'] = worker_replica_spec

        tf_job_spec = {}
        tf_job_spec['kind'] = 'TFJob'
        tf_job_spec['apiVersion'] = 'kubeflow.org/' + TFJobHandler.VERSION
        tf_job_spec['metadata'] = client.V1ObjectMeta(generate_name=task.name)
        tf_job_spec['spec'] = spec

        tf_job = self.api.create_namespaced_custom_object(
            TFJobHandler.GROUP, TFJobHandler.VERSION, self.namespace, TFJobHandler.PLURAL, tf_job_spec
        )
        return tf_job['metadata']['name']

    def status(self, name: str, store: Store) -> Status:
        """tf-operator seems to have no way to support querying status, instead, find out its pods' status

        :param name: The task name
        :type name: Task
        :param store: The job store
        :type store: Store
        :return: A status
        :rtype: Status
        """

        resource_id = store.get(name)[Store.RESOURCE_ID]
        statuses = [p.status for p in self.get_pods(resource_id)]

        # The idea here is that every single tfjob pod has to be done
        if not all([s.phase in TFJobHandler.TERMINAL_PHASES for s in statuses]):
            return Status(StatusType.RUNNING, None)

        status = statuses[0]
        return status

    def kill(self, name, store: Store) -> None:
        """Kill a TFJob

        :param name: The job name
        :type name: str
        :param store: The job store
        :type store: Store
        :return: None
        """
        delete_options = client.V1DeleteOptions(api_version=TFJobHandler.VERSION, propagation_policy="Background")
        resource_id = store.get(name)[Store.RESOURCE_ID]
        return self.api.delete_namespaced_custom_object(
            TFJobHandler.GROUP,
            TFJobHandler.VERSION,
            self.namespace,
            TFJobHandler.PLURAL,
            resource_id,
            body=delete_options,
        )



from typing import Optional, List, Any
import pymongo
from odin.store import Cache, Store, Dict, register_cache_backend, register_store_backend


@register_cache_backend('mongo')
class MongoCache(Cache):
    """A key-value cache backed by a Mongodb."""

    COLL = 'hashes'
    IN = 'inputs'
    OUT = 'outputs'
    ID = '_id'

    def __init__(
        self,
        host: str,
        user: Optional[str] = None,
        passwd: Optional[str] = None,
        db: str = 'jobs_db',
        port: int = pymongo.MongoClient.PORT,
        **kwargs,
    ):
        """Connect to the mongodb backend.

        :param host: The location of the db.
        :param user: The username to log in with.
        :param passwd: The password to use.
        :param db: The database to use.
        :param port: The port to connect on.
        :raises Exception: If connection to mongo fails.
        :raises ServerSelectionTimeoutError: If the connection times out.
        """
        super().__init__()
        self.dbhost = host
        if user and passwd:
            uri = f"mongodb://{user}:{passwd}@{host}:{port}/{db}"
            client = pymongo.MongoClient(uri)
        else:
            client = pymongo.MongoClient(host, port)
        if client is None:
            raise Exception(f"Cannot connect to mongo: [{host}:{port}] as user: [{user}]")
        try:
            self.db = client.get_database(db)
        except pymongo.errors.ServerSelectionTimeoutError:
            raise Exception(f"Connot get db from mongo: [{host}:{port}], connections timed out")
        # Creating an index in Mongo is Idempotent
        self.db[MongoCache.COLL].create_index([(MongoCache.IN, pymongo.ASCENDING)], unique=True)

    def __setitem__(self, key: str, value: Any) -> None:
        """Add to the cache, if the key exists overwrite it.

        :param key: The key to add to the cache.
        :param value: The value that goes with the key
        """
        self.db[MongoCache.COLL].replace_one(
            {MongoCache.IN: key}, {MongoCache.IN: key, MongoCache.OUT: value}, upsert=True
        )

    def __getitem__(self, key: str) -> Any:
        """Get data from the cache. Returns None if key is missing.

        :param key: The key to look for.
        :returns: The value associated with the key.
        """
        hit = self.db[MongoCache.COLL].find_one({MongoCache.IN: key})
        return hit[MongoCache.OUT] if hit else None

    def __delitem__(self, key: str) -> None:
        """Delete a key-value pair from the cache.

        :param key: The key to remove.
        """
        self.db[MongoCache.COLL].delete_one({MongoCache.IN: key})

    def keys(self) -> List[str]:
        """Get the keys in the cache."""
        return [doc[MongoCache.IN] for doc in self.db[MongoCache.COLL].find({}, {MongoCache.IN: 1, MongoCache.ID: 0})]


@register_store_backend('mongo')
class MongoStore(Store):
    """MongoDB backed `Store` implementation

    This implementation fulfills the `Store` interface backed by a MongoDB
    """

    def __init__(self, host, user, passwd, db='jobs_db', port=pymongo.MongoClient.PORT, **kwargs):
        """A MongoStore is a Store implemented using MongoDB
        """
        super().__init__()
        self.dbhost = host
        if user and passwd:
            uri = f"mongodb://{user}:{passwd}@{host}:{port}/{db}"
            client = pymongo.MongoClient(uri)
        else:
            client = pymongo.MongoClient(host, port)
        if client is None:
            error_str = f"cannot connect to mongo: [{host}:{port}] as user: [{user}]"
            raise Exception(error_str)
        try:
            self.db = client.get_database(db)

        except pymongo.errors.ServerSelectionTimeoutError:
            raise Exception(f"cannot get db from mongo: [{host}:{port}], connection timed out")

    def _get_if(self, job_id: str):
        return self.db[MongoStore.JOBS].find_one({Store.PIPE_ID: job_id})

    def get(self, job_id: str) -> Dict:
        """This gives back the result of the job store for this entry

        :param job_id: This is a unique ID for this job
        :raises KeyError: If the job is not in the database.
        :return: A dictionary containing the output user data
        """
        result = self._get_if(job_id)
        if result is None:
            raise KeyError(f"No job {job_id} found in jobs DB")
        return result

    def get_parent(self, job_str: str) -> Dict:
        """Get job results from the parent job

        :param job_str: This job
        :return: The parent job
        """
        return self.get(self.get(job_str)[Store.PARENT])

    def get_previous(self, job_str: str) -> List[Dict]:
        """Get all job results that can before this job

        :param job_str: This job
        :return: All previous
        """
        parent = self.get_parent(job_str)
        return [self.get(prev) for prev in parent[Store.EXECUTED]]

    def _set(self, value: Dict) -> None:
        """This updates the job store for this entry

        :param value:
        """
        self.db[MongoStore.JOBS].replace_one({Store.PIPE_ID: value[Store.PIPE_ID]}, value, upsert=True)

    def exists(self, job_id: str) -> bool:
        """Check if there is a job in the database with this id

        :param job_id:
        :return:
        """
        return self._get_if(job_id) is not None

    def remove(self, job_id: str) -> bool:
        """Delete a job from the database

        :param job_id:
        :return: Did the removal succeed.
        """
        result = self._get_if(job_id)
        if result is None:
            return False

        db_id = result['_id']
        self.db[MongoStore.JOBS].remove({'_id': db_id})
        return True

    def parents_like(self, pattern: str) -> List[str]:
        """Get all the parent jobs that match some pattern

        :param pattern: A pattern to match
        :return: Any parents that match this
        { <field>: { $regex: /pattern/<options> } }
        """
        parents = self.db[MongoStore.JOBS].find({Store.PIPE_ID: {'$regex': pattern}, Store.PARENT: None})

        return [p[Store.PIPE_ID] for p in parents]

    def children_like(self, pattern: str) -> List[str]:
        """Get all jobs that match some pattern

        :param pattern: A pattern to match
        :return: Any jobs that match this
        """
        children = self.db[MongoStore.JOBS].find({Store.PIPE_ID: {'$regex': pattern}, Store.PARENT: {"$ne": None}})

        return [c[Store.PIPE_ID] for c in children]

    def is_a_child(self, child: str) -> bool:
        """Is the resource a child of a job

        :param child: The name of the child
        :return: `True` if its a child
        """
        children = self.children_like(child)
        return bool(children)



"""Websocket client to get the k8s events that act on some resource."""

import json
import asyncio
import argparse
import websockets
from odin import LOGGER, APIField, APIStatus
from odin.k8s import Event
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME, HttpClient
from odin.utils.formatting import print_table


async def request_events(url: str, resource: str, namespace: str = 'default') -> None:
    """Get k8s events for some resource.

    :param url: The location of the server
    :param resource: The name of the resource you are asking about.
    :param namespace: The namespace of the resource you are asking about.
    """
    async with websockets.connect(url) as websocket:
        await websocket.send(
            json.dumps({APIField.COMMAND: 'EVENTS', APIField.REQUEST: {'resource': resource, 'namespace': namespace}})
        )
        resp = json.loads(await websocket.recv())
        if resp[APIField.STATUS] == APIStatus.ERROR:
            LOGGER.error(resp)
            return
        if resp[APIField.STATUS] == APIStatus.OK:
            rows = [Event(**r) for r in resp[APIField.RESPONSE]]
            print_table(rows)


def _result2event(result):
    event = Event(
        type=result.get('eventType'),
        reason=result.get('reason'),
        source=result.get('source'),
        message=result.get('message'),
        timestamp=result.get('timestamp'),
    )
    return event


def request_events_http(url: str, resource: str) -> None:
    """Get events for a resource over HTTP

    :param url: The base URL
    :param resource: The resource ID
    """
    results = HttpClient(url).request_events(resource)
    events = [_result2event(result) for result in results['events']]
    print_table(events)


def main():
    """Websocket client for pinging odin."""
    parser = argparse.ArgumentParser(description='Websocket-based health check')
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument(
        '--scheme',
        choices={'wss', 'ws', 'http', 'https'},
        default=ODIN_SCHEME,
        help='Connection protocol, use `http` for REST, use `wss` for remote connections and `ws` for localhost',
    )
    parser.add_argument('resource', help="The name of the resource to describe")
    parser.add_argument('--namespace', default='default', help="The namespace the resource lives in.")
    args = parser.parse_args()
    url = f'{args.scheme}://{args.host}:{args.port}'

    if args.scheme.startswith('ws'):
        asyncio.get_event_loop().run_until_complete(request_events(url, args.resource, args.namespace))
    else:
        request_events_http(url, args.resource)


if __name__ == "__main__":
    main()



"""Defines a resource handler for Pods"""

from typing import List
from kubernetes import client
from odin.store import Store
from odin.k8s import Task, Status, ResourceHandler, task_to_pod_spec, register_resource_handler


@register_resource_handler
class PodHandler(ResourceHandler):
    """A resource handler for Pods"""

    NAME = 'Pod'

    @property
    def kind(self) -> str:
        """Give back the k8s "kind"

        :return: kind
        :rtype: str
        """
        return PodHandler.NAME

    def submit(self, task: Task) -> str:
        """Submit a pod to run

        :param task: A task/pod
        :type task: Task
        :return: A string name
        :rtype str
        """
        secrets = self._reference_secrets(task)
        configmaps = self._generate_configmaps(task)
        pod_spec = task_to_pod_spec(task, secrets=secrets, configmaps=configmaps)
        metadata = client.V1ObjectMeta(name=task.name)

        pod = client.V1Pod(metadata=metadata, spec=pod_spec)
        self.core_api.create_namespaced_pod(body=pod, namespace=self.namespace)
        return task.name

    def get_api(self) -> object:
        """Get the (core) API for this resource handler

        :return: The core API
        :rtype: object
        """
        return self.core_api

    def get_pods(self, name: str) -> List[client.models.v1_pod.V1Pod]:
        """Get all pods for this Task

        :param: name A name
        :type: str
        :return: All pods
        :rtype: List[client.models.v1_pod.V1Pod]
        """
        results = self.core_api.read_namespaced_pod(name, self.namespace)

        return [results] if results else []

    def status(self, name: str, store: Store) -> Status:
        """Get status for this pod

        :param name: The name
        :type name: str
        :param store: The job store
        :type store: Store
        :return: A status
        :rtype: Status
        """
        pod_status = self.core_api.read_namespaced_pod_status(name=name, namespace=self.namespace).status
        return pod_status

    def kill(self, name: str, store: Store) -> None:
        """Kill a pod

        :param name: pod name
        :type name: str
        :param store: A store
        :type store: Store
        """
        self.core_api.delete_namespaced_pod(name=name, namespace=self.namespace)



"""Cleanup odin jobs
"""
import os
import shutil
import argparse
from collections import namedtuple
from itertools import chain
from typing import List, Optional, Set

from baseline.utils import exporter, read_config_stream
from mead.utils import convert_path
from odin import LOGGER
from odin.k8s import KubernetesTaskManager, TaskManager
from odin.store import Store, create_store_backend
from odin.utils.formatting import print_table


__all__ = []
export = exporter(__all__)


Cleaned = namedtuple('Cleaned', 'task_id cleaned_from_k8s purged_from_db removed_from_fs')


def done(job: str, completed: Set[str]) -> str:
    """Convert set membership into `Yes` for `No`.

    :param job: The job to check if it was acted on
    :param completed: The jobs acted on,
    :returns: Yes or No
    """
    return 'Yes' if job in completed else 'No'


@export
def cleanup(
    work: str,
    store: Store,
    sched: Optional[TaskManager] = None,
    purge_db: bool = False,
    purge_fs: bool = False,
    data_dir: Optional[str] = None,
) -> List[Cleaned]:
    """Cleanup a job

    :param work: The name of the job
    :param store: The job store
    :param sched: The scheduler used to kill jobs.
    :param purge_db: Should the pipeline also be removed from the job db.
    :param purge_fs: Should the pipeline also be removed from the file system.
    :param data_dir: A directory that combined with work is where all artifacts produced by the pipeline live.

    :returns: The list jobs and if it was removed from k8s or the job db.
    """
    if not work:
        return []
    sched = sched if sched else KubernetesTaskManager(store)
    parent_details = store.get(work)
    children = list(chain(parent_details[Store.EXECUTED], parent_details[Store.EXECUTING]))
    cleaned = set()
    purged = set()
    removed = set()
    if purge_fs:
        if data_dir is None:
            LOGGER.warning("Requested removal from the file system but no data directory provided.")
        else:
            shutil.rmtree(os.path.join(data_dir, work), ignore_errors=True)
            removed = set(chain([work], children))
    for job in children:
        try:
            sched.kill(job)
            cleaned.add(job)
        except:  # pylint: disable=bare-except
            pass
        if purge_db:
            if store.remove(job):
                purged.add(job)
    # Remove the work entry from the db last so if there is an error before hand we can still use the db entry.
    if purge_db:
        if store.remove(work):
            purged.add(work)
    return [Cleaned(j, done(j, cleaned), done(j, purged), done(j, removed)) for j in chain([work], children)]


def main():
    """Clean up a job
    """
    parser = argparse.ArgumentParser(description='Clean up a job')
    parser.add_argument('work', help='Job')
    parser.add_argument('--cred', help='cred file', type=convert_path, required=True)
    parser.add_argument('--db', action='store_true', help="Also remove from the jobs db")
    parser.add_argument('--fs', action='store_true', help="Also remove from the filesystem")
    parser.add_argument('--data_dir', help="The root of where data is saved.")

    args = parser.parse_args()

    cred_params = read_config_stream(args.cred)
    store = create_store_backend(**cred_params['jobs_db'])

    cleaned = cleanup(args.work, store, purge_db=args.db, purge_fs=args.fs, data_dir=args.data_dir)
    print("Results of this request:")
    print_table(cleaned)


if __name__ == "__main__":
    main()



"""Serve pipelines over websockets
"""
import os
import json
import logging
import argparse
import asyncio
import traceback
from datetime import datetime, date
from typing import Dict, Any, Optional
from functools import partial
import websockets
import git
from bson.json_util import dumps as bson_dumps
from eight_mile.utils import read_config_stream
from mead.utils import convert_path
from odin import ODIN_LOGO, APIField, APIStatus
from odin.cleanup import cleanup
from odin.core import read_pipeline_config
from odin.generate import generate_pipeline
from odin.executor import Executor
from odin.status import get_status
from odin.store import create_store_backend, create_cache_backend
from odin.k8s import KubernetesTaskManager, KF_MODULES, ELASTIC_MODULES, CORE_MODULES

LOGGER = logging.getLogger('odin')

STORE = None
CACHE = None
ROOT_PATH = None
DATA_PATH = None


def seralize(obj: Any) -> str:
    """Callable on non-serialize able objects, handles datatimes.

    :param obj: The unserialzable object.
    :raises TypeError: When the object can't be serialized.
    :returns: The serialized object.
    """
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError(f"Type {type(obj)} not JSON serializable")


async def send_with_extra(websocket: 'Websocket', extras: Dict, data: Dict) -> None:
    """Send data with extras added encoded as a json string over websocket.

    :param websocket: The connection to send the data over
    :param extras: Extra data to add into the response before you send it
    :param data: The response
    """
    data.update(extras)
    await websocket.send(json.dumps(data, default=seralize))


async def handle_request(websocket, _) -> None:  # pylint: disable=too-many-branches,too-many-statements,too-many-locals
    """Async handler for a websocket, runs the pipeline

    :param websocket: A handle to the socket
    :param _: ?
    """
    global ROOT_PATH
    global DATA_PATH
    request = await websocket.recv()
    LOGGER.info(request)

    try:
        request = json.loads(request)
        cmd = request.pop(APIField.COMMAND)
        work = request.pop(APIField.REQUEST)
        send = partial(send_with_extra, websocket, request)

        if cmd == 'START':
            work = os.path.normpath(work)
            work_path = os.path.join(ROOT_PATH, work)
            context, tasks = read_pipeline_config(work_path, ROOT_PATH, DATA_PATH)
            LOGGER.info(context)
            pipe = Executor(STORE, cache=CACHE, modules=MODULES)
            pipe_id = context['PIPE_ID']
            try:
                repos = git.Repo(ROOT_PATH)
                rev_ver = str(git.repo.fun.rev_parse(repos, 'HEAD'))
            except Exception:
                LOGGER.warning("Failed to get revision version")
                rev_ver = None

            await send({APIField.STATUS: APIStatus.OK, APIField.RESPONSE: pipe_id})
            async for log in pipe.run(pipe_id, os.path.basename(work_path), rev_ver, tasks):
                try:
                    await send({APIField.STATUS: APIStatus.OK, APIField.RESPONSE: log})
                except websockets.exceptions.ConnectionClosed:
                    continue
            try:
                await send({APIField.STATUS: APIStatus.END, APIField.RESPONSE: pipe_id})
            except websockets.exceptions.ConnectionClosed:
                return
        elif cmd == 'SHOW':
            pipeline_loc = os.path.join(ROOT_PATH, work)
            try:
                defs = {f: open(os.path.join(pipeline_loc, f)).read() for f in os.listdir(pipeline_loc)}
                await send({APIField.STATUS: APIStatus.OK, APIField.RESPONSE: defs})
            except FileNotFoundError as exc:
                LOGGER.error(exc)
                await send({APIField.STATUS: APIStatus.ERROR, APIField.RESPONSE: str(exc)})
        elif cmd == 'GENERATE':
            try:
                pipeline_name = generate_pipeline(ROOT_PATH, **work)
                await send({APIField.STATUS: APIStatus.OK, APIField.RESPONSE: pipeline_name})
            except FileExistsError as exc:
                LOGGER.error(exc)
                await send({APIField.STATUS: APIStatus.ERROR, APIField.RESPONSE: str(exc)})
        elif cmd == 'CLEANUP':
            cleaned = [c._asdict() for c in cleanup(**work, store=STORE, data_dir=DATA_PATH)]
            await send({APIField.STATUS: APIStatus.OK, APIField.RESPONSE: cleaned})
        elif cmd == 'STATUS':
            work = STORE.parents_like(work)
            results = [
                {'pipeline_status': {**pipe._asdict()}, 'task_statuses': [r._asdict() for r in rows]}
                for pipe, rows in (get_status(w, STORE) for w in work)
            ]
            await send({APIField.STATUS: APIStatus.OK, APIField.RESPONSE: results})
        elif cmd == 'PING':
            response = f"PONG {work}"
            await send({APIField.STATUS: APIStatus.OK, APIField.RESPONSE: response})
        elif cmd == 'EVENTS':
            task_mgr = KubernetesTaskManager(STORE, modules=MODULES)
            await send(
                {
                    APIField.STATUS: APIStatus.OK,
                    APIField.RESPONSE: [e._asdict() for e in task_mgr.get_events(name=work.get('resource'))],
                }
            )
        elif cmd == 'DATA':
            obj = STORE.get(work.get('resource'))
            obj = json.loads(bson_dumps({'success': True, 'jobs': obj}))
            await send({APIField.STATUS: APIStatus.OK, APIField.RESPONSE: obj})
        elif cmd == 'LOGS':
            task_mgr = KubernetesTaskManager(STORE, modules=MODULES)
            name = work.pop('resource')
            work.pop('namespace')
            work['name'] = name
            if work.pop('follow', False):
                async for log in task_mgr.follow_logs(name):
                    try:
                        await send({APIField.STATUS: APIStatus.OK, APIField.RESPONSE: log})
                    except websockets.exceptions.ConnectionClosed:
                        return
            else:
                await send({APIField.STATUS: APIStatus.OK, APIField.RESPONSE: task_mgr.get_logs(**work)})
            await send({APIField.STATUS: APIStatus.END, APIField.RESPONSE: 'LOGS'})
        else:
            await send({APIField.STATUS: APIStatus.ERROR, APIField.RESPONSE: f"{cmd} not found."})
    except Exception as exc:
        LOGGER.error(traceback.format_exc())
        # Because this error can happen before send is bound and we send the request back anyway there is no need to
        # do the merge and send code.
        await websocket.send(json.dumps({APIField.STATUS: APIStatus.ERROR, APIField.RESPONSE: str(exc)}))


def get_db_config(cred: Optional[str]) -> Dict:
    """

    :param cred:
    :return:
    """
    if cred:
        cred_params = read_config_stream(cred)['jobs_db']

    else:
        cred_params = {}
        cred_params['backend'] = os.environ.get("ODIN_JOBS_BACKEND", "postgres")
        cred_params['host'] = os.environ.get("SQL_HOST", "127.0.0.1")
        cred_params['port'] = os.environ.get("DB_PORT", 5432)
        cred_params['user'] = os.environ.get("DB_USER")
        cred_params['passwd'] = os.environ.get("DB_PASS")
    cred_params['db'] = os.environ.get("DB_NAME", "jobs_db")
    return cred_params


def main():
    """Launch a websocket server that will run forever and serve pipelines
    """
    global STORE
    global CACHE
    global ROOT_PATH
    global DATA_PATH
    global MODULES
    parser = argparse.ArgumentParser(description='Websocket-based Pipeline scheduler')
    parser.add_argument('--root_path', help='Root directory', type=convert_path, required=True)
    parser.add_argument('--host', default='0.0.0.0', type=str)
    parser.add_argument('--port', default='30000')
    parser.add_argument('--cred', help='cred file', type=convert_path)
    parser.add_argument('--data_path', help='data directory')
    parser.add_argument('--modules', default='all', choices=['all', 'kf', 'elastic', 'core'])
    args = parser.parse_args()

    MODULES = CORE_MODULES
    if args.modules in ['kf', 'all']:
        MODULES += KF_MODULES
    if args.modules in ['elastic', 'all']:
        MODULES = ELASTIC_MODULES + CORE_MODULES
    config_params = get_db_config(args.cred)
    STORE = create_store_backend(**config_params)
    CACHE = create_cache_backend(**config_params)
    ROOT_PATH = os.path.normpath(args.root_path)
    args.data_path = args.data_path if args.data_path is not None else args.root_path
    DATA_PATH = os.path.normpath(args.data_path)

    start_server = websockets.serve(handle_request, args.host, args.port)
    asyncio.get_event_loop().run_until_complete(start_server)
    LOGGER.info(ODIN_LOGO)
    LOGGER.info("Ready to serve.")
    asyncio.get_event_loop().run_forever()


if __name__ == "__main__":
    main()



"""Client code specific constants."""
import os
import requests
from typing import Dict
from odin.utils.auth import _authenticate

ODIN_URL = os.environ.get('ODIN_URL', 'localhost')
ODIN_PORT = os.environ.get('ODIN_PORT', 9003)
ODIN_SCHEME = os.environ.get('ODIN_SCHEME', 'https')


def encode_path(path: str) -> str:
    """Encode a path from `/` to `__`

    :param path: A path to encode
    :return: An encoded vector
    """
    vec = []
    while path:
        head, tail = os.path.split(path)
        vec.append(tail)
        path = head
    return '__'.join(filter(lambda x: x, vec[::-1]))


class HttpClient:
    def __init__(self, url=None, host=None, port=None, scheme=None, username=None, password=None, jwt_token=None):
        if url is None:
            if host is None:
                host = ODIN_URL
            if port is None:
                port = ODIN_PORT
            if scheme is None:
                scheme = ODIN_SCHEME
            self.url = f'{scheme}://{host}:{port}'
        else:
            self.url = url

        username = username
        password = password
        if username is not None and password is not None:
            self.jwt_token = _authenticate(self.url, username, password)
        else:
            self.jwt_token = jwt_token

    def schedule_pipeline(self, work: str, context: Dict={}) -> Dict:
        """Request the status over HTTP
        :param url: the base URL
        :param work: The pipeline ID
        :param context: Optional context values
        """
        job = encode_path(work)

        response = requests.post(
            f'{self.url}/v1/pipelines',
            headers={'Authorization': f'Bearer {self.jwt_token}'},
            json={"pipeline": {"job": job}, "context": context},
        )
        if response.status_code == 401:
            raise ValueError("Invalid login")
        results = response.json()
        return results

    def create_job(self, name: str) -> Dict:

        """Request the server makes a new job.

        :param url: Base url of the remote odin server
        :param name: The name of the job you want to create
        """
        job = encode_path(name)
        response = requests.post(
            f"{self.url}/v1/jobs", headers={"Authorization": f"Bearer {self.jwt_token}"}, json={"job": {"name": job}}
        )
        if response.status_code == 401:
            raise ValueError("Invalid Login")
        results = response.json()
        return results

    def request_job(self, name: str) -> Dict:
        name = name.replace('/', '__')
        response = requests.get(f'{self.url}/v1/jobs/{name}')
        results = response.json()
        return results

    def request_events(self, resource: str) -> Dict:
        """Get events for a resource over HTTP

        :param url: The base URL
        :param resource: The resource ID
        """
        response = requests.get(f'{self.url}/v1/resources/{resource}/events')
        if response.status_code == 401:
            raise ValueError("Invalid login")
        results = response.json()
        return results

    def request_data(self, resource: str) -> Dict:
        """Get data for a resource over HTTP

        :param url: The base URL
        :param resource: The resource ID
        """
        response = requests.get(f'{self.url}/v1/resources/{resource}/data')
        results = response.json()
        return results

    def request_cluster_hw_status(self) -> Dict:
        """Request the status over HTTP
        """
        response = requests.get(f'{self.url}/v1/nodes')
        nodes = response.json()['nodes']
        return nodes

    def push_file(self, job: str, file_name: str, file_contents: str) -> Dict:
        """Push a file to update a pipeline.

        :param job: The job definition that will be updated
        :param file_name: The name to save the file as on the remove server
        :param file_contents: The content of the file we want to upload
        """
        job = encode_path(job)
        response = requests.post(
            f'{self.url}/v1/jobs/{job}/files/{file_name}',
            data=file_contents,
            headers={'Content-Type': 'text/plain', 'Authorization': f'Bearer {self.jwt_token}'},
        )
        if response.status_code == 401:
            raise ValueError("Invalid login")
        results = response.json()
        return results

    def delete_pipeline(self, work: str, purge_db: bool = False, purge_fs: bool = False) -> Dict:
        """Request delete pipeline
        :param url: the base URL
        :param work: The pipeline ID
        :param purge_db: Should we delete the pipeline from the jobs db too?
        :param purge_fs: Should we remove pipeline file system artifacts?
        """

        response = requests.delete(
            f'{self.url}/v1/pipelines/{work}',
            headers={'Authorization': f'Bearer {self.jwt_token}'},
            params={'db': purge_db, 'fs': purge_fs},
        )
        if response.status_code == 401:
            raise ValueError("Invalid login")
        results = response.json()
        return results

    def request_status(self, work: str) -> Dict:
        """Request the status over HTTP
        :param url: the base URL
        :param work: The pipeline ID
        :param columns: A set of columns to include in the output
        :param all_cols: Should we just show all columns, If true then columns in ignored
        """
        response = requests.get(f'{self.url}/v1/pipelines?q={work}')
        results = response.json()['pipelines']
        return results

    def request_users(self):
        response = requests.get(f'{self.url}/v1/users')
        results = response.json()['users']
        return results

    def request_logs(self, pod: str, namespace='default', **kwargs) -> str:
        """This function really shouldnt be provided here -- forget you saw this

        Using the logs directly from k8s isnt desirable, we should add this to the HTTP tier, but if you really
        need it, this function should get results back
        :param work:
        :return:
        """
        from kubernetes import client, config

        try:
            config.load_incluster_config()
        except config.config_exception.ConfigException:
            config.load_kube_config()
        api = client.CoreV1Api()
        logs = api.read_namespaced_pod_log(pod, namespace=namespace, **kwargs)
        return logs

    def app_info(self):
        """Get information about run configuration

        :return:
        """
        response = requests.get(f'{self.url}/v1/app')
        return response.json()



"""A demo chore that sends a message to slack, used to test/demonstrate chore plugins."""

import requests
from odin.chores import register_chore


@register_chore('demo-chore')
def slack_webhook(webhook: str) -> None:
    """Send a message on slack.

    :param webhook: The url to send the message to.
    """
    requests.post(webhook, json={"text": "This is a demo slack message sent with a chore addon"})



"""Defines a resource handler for multi-worker PyTorchJobs"""

from typing import List
from kubernetes import client
from odin.store import Store
from odin.k8s import (
    Task,
    Status,
    ResourceHandler,
    StatusType,
    task_to_pod_spec,
    json_to_selector,
    register_resource_handler,
)


@register_resource_handler(aliases=['pytjob'])
class PyTorchJobHandler(ResourceHandler):
    """Resource handler for multi-worker PyTorchJobs"""

    NAME = "PyTorchJob"
    GROUP = "kubeflow.org"
    PLURAL = "pytorchjobs"
    VERSION = "v1"
    # This label selector was changed, we are using the new format
    SELECTOR = "pytorch-job-name"
    GROUP_KEY = "group-name"

    @property
    def kind(self) -> str:
        """Give back the k8s "kind"

        :return: kind
        :rtype: str
        """
        return PyTorchJobHandler.NAME

    def __init__(self, namespace: str):
        """Create a resource handler with resources in the given namespace

        :param namespace: A namespace for resources
        :type namespace: str
        """
        super().__init__(namespace)
        self.api = client.CustomObjectsApi()

    def get_api(self) -> object:
        """Get the API for this resource handler

        :return: An API for this handler
        :rtype: object
        """
        return self.api

    def submit(self, task: Task) -> str:
        """Submit a multi-worker PyTorchJob Task

        :param task: The task definition
        :type task: Task
        :return: A string handle name
        :rtype: str
        """
        secrets = self._reference_secrets(task)
        configmaps = self._generate_configmaps(task)
        task.num_gpus = 1
        pod_spec = task_to_pod_spec(task, container_name="pytorch", secrets=secrets, configmaps=configmaps)
        template_metadata = client.V1ObjectMeta(name=task.name)

        template = client.V1PodTemplateSpec(metadata=template_metadata, spec=pod_spec)

        worker_replica_spec = {}
        worker_replica_spec['replicas'] = task.num_workers
        worker_replica_spec['restartPolicy'] = PyTorchJobHandler.RESTART_NEVER
        worker_replica_spec['template'] = template

        master_replica_spec = {}
        master_replica_spec['replicas'] = 1
        master_replica_spec['restartPolicy'] = PyTorchJobHandler.RESTART_NEVER
        master_replica_spec['template'] = template

        spec = {}
        spec['pytorchReplicaSpecs'] = {}
        spec['pytorchReplicaSpecs']['Master'] = master_replica_spec
        spec['pytorchReplicaSpecs']['Worker'] = worker_replica_spec

        pytorch_job_spec = {}
        pytorch_job_spec['kind'] = "PyTorchJob"
        pytorch_job_spec['apiVersion'] = 'kubeflow.org/' + PyTorchJobHandler.VERSION
        pytorch_job_spec['metadata'] = client.V1ObjectMeta(generate_name=task.name)
        pytorch_job_spec['spec'] = spec

        pytorch_job = self.api.create_namespaced_custom_object(
            PyTorchJobHandler.GROUP,
            PyTorchJobHandler.VERSION,
            self.namespace,
            PyTorchJobHandler.PLURAL,
            pytorch_job_spec,
        )
        return pytorch_job['metadata']['name']

    def get_pods(self, name: str) -> List[client.models.v1_pod.V1Pod]:
        """Find all the pods contained in a pytorch job.

        :param name: The name of job (the jobs db resource id)
        :type name: str
        :returns: The pods in the job
        :rtype: List[client.models.v1_pod.V1Pod]
        """
        try:
            selector = json_to_selector(
                {PyTorchJobHandler.SELECTOR: name, PyTorchJobHandler.GROUP_KEY: PyTorchJobHandler.GROUP}
            )
            return self.core_api.list_namespaced_pod(self.namespace, label_selector=selector).items
        except client.rest.ApiException:
            return []

    def status(self, name: str, store: Store) -> Status:
        """Find out its pods' status

        :param name: A task name
        :type name: str
        :param store: A job store
        :type store: Store
        :return: A job status
        :rtype: Status
        """

        resource_id = store.get(name)[Store.RESOURCE_ID]
        statuses = [p.status for p in self.get_pods(resource_id)]

        # The idea here is that every single pod has to be done
        if not all([s.phase in PyTorchJobHandler.TERMINAL_PHASES for s in statuses]):
            return Status(StatusType.RUNNING, None)

        status = statuses[0]
        return status

    def kill(self, name: str, store: Store) -> None:
        """Kill a multi-worker PyTorch task with given name

        :param name: The task to kill
        :type name: str
        :param store: The jobs store
        :type store: Store
        :return: None
        """
        delete_options = client.V1DeleteOptions(api_version=PyTorchJobHandler.VERSION, propagation_policy="Background")
        resource_id = store.get(name)[Store.RESOURCE_ID]
        return self.api.delete_namespaced_custom_object(
            PyTorchJobHandler.GROUP,
            PyTorchJobHandler.VERSION,
            self.namespace,
            PyTorchJobHandler.PLURAL,
            resource_id,
            body=delete_options,
        )



"""Client to run odin jobs"""
import argparse
from getpass import getuser
import os
import asyncio
import json
import signal
from typing import Dict
import websockets
from mead.utils import parse_and_merge_overrides
from odin import LOGGER, APIField, APIStatus
from odin.client import ODIN_URL, ODIN_PORT, ODIN_SCHEME, HttpClient
from odin.utils.auth import get_jwt_token


async def schedule_pipeline(ws, work) -> None:
    """Use async to open a connection to serve.py and launch work

    Blocks until the job completes (and websocket stays open)
    """
    async with websockets.connect(ws) as websocket:
        await websocket.send(json.dumps({APIField.COMMAND: 'START', APIField.REQUEST: work}))

        result = json.loads(await websocket.recv())
        while result[APIField.STATUS] != APIStatus.END:
            if result[APIField.STATUS] == APIStatus.ERROR:
                LOGGER.error(result)
                return

            if result[APIField.RESPONSE].startswith('PIPE_ID'):
                pipe_id = result.split(' ')[-1]
                LOGGER.info('Started %s', pipe_id)
            else:
                LOGGER.info(result[APIField.RESPONSE])
            result = json.loads(await websocket.recv())


def schedule_pipeline_http(url: str, jwt_token: str, work: str, context: Dict) -> None:
    """Request the status over HTTP
    :param url: the base URL
    :param jwt_token: The JWT token representing this authentication
    :param work: The pipeline ID
    """

    results = HttpClient(url, jwt_token=jwt_token).schedule_pipeline(work, context)
    print(json.dumps(results))


def main():
    """Use `asyncio` to connect to a websocket and request a pipeline, wait.
    """
    signal.signal(signal.SIGINT, lambda *args, **kwargs: exit(0))

    parser = argparse.ArgumentParser(description='HTTP or Websocket-based Pipeline scheduler')
    parser.add_argument('work', help='Job')
    parser.add_argument('--host', default=ODIN_URL, type=str)
    parser.add_argument('--port', default=ODIN_PORT)
    parser.add_argument('--token', help="File where JWT token can reside", default=os.path.expanduser("~/.odin.token"))
    parser.add_argument('--username', '-u', help="Username", default=getuser())
    parser.add_argument('--password', '-p', help="Password")
    parser.add_argument(
        '--scheme',
        choices={'http', 'wss', 'ws', 'https'},
        default=ODIN_SCHEME,
        help='Connection protocol, use `http` for REST, use `wss` for remote connections and `ws` for localhost',
    )

    args, overrides = parser.parse_known_args()
    context = parse_and_merge_overrides({}, overrides, pre='x')


    url = f'{args.scheme}://{args.host}:{args.port}'

    if args.scheme.startswith('ws'):
        if context:
            LOGGER.warning("Context is ignored by web-socket tier")
        asyncio.get_event_loop().run_until_complete(schedule_pipeline(url, args.work))
    else:
        jwt_token = get_jwt_token(url, args.token, args.username, args.password)
        try:
            schedule_pipeline_http(url, jwt_token, args.work, context)
        except ValueError:
            # Try deleting the token file and start again
            if os.path.exists(args.token):
                os.remove(args.token)
                jwt_token = get_jwt_token(url, args.token, args.username, args.password)
                schedule_pipeline_http(url, jwt_token, args.work, context)


if __name__ == '__main__':
    main()



"""Job store API
"""
import re
import os
from typing import Dict, List, Optional, Any
import logging
from eight_mile.utils import read_config_file, register, optional_params
from baseline.utils import import_user_module

# Same constants as defined in odin.k8s, repeated here to avoid circular import
ODIN_TASK_ENV = "ODIN_TASK_ID"
ODIN_CRED_ENV = "ODIN_CRED"


LOGGER = logging.getLogger('odin')


class Store:
    """Base interface for a job store
    """

    JOBS = 'jobs'
    STATUS = 'status'
    PARENT = 'parent'
    EXECUTED = 'executed'
    EXECUTING = 'executing'
    REV_VER = 'version'
    WAITING = 'waiting'
    JOB_NAME = 'job'
    PIPE_ID = 'label'
    ERROR_MESSAGE = 'error_message'
    REQUEST_EARLY_EXIT = 'request_early_exit'
    DONE = 'DONE'
    COMPLETION_TIME = 'completion_time'
    SUBMIT_TIME = 'submit_time'
    RESOURCE_ID = 'resource_id'
    RESOURCE_TYPE = 'resource_type'
    CACHED = 'Cached'

    def __init__(self):
        """A Store is an abstract interface for talking to a job database
        """

    def get(self, job_id: str) -> Dict:
        """This gives back the result of the job store for this entry

        :param job_id: This is a unique ID for this job
        :return: A dictionary containing the output user data
        """

    def set(self, value: Dict) -> None:
        """This updates the job store for this entry

        This Dict must include a `label` key
        :param value: An object to store for this job
        """
        self._set_preconditions(value)
        self._set(value)

    def _set(self, value: Dict):
        """Actual setter function

        :param value:
        :return:
        """

    def exists(self, job_id: str) -> bool:
        """Check if there is a job in the database with this id

        :param job_id:
        :return:
        """

    def remove(self, job_id: str) -> bool:
        """Delete a job from the database

        :param job_id:
        :return: If the job could be removed
        """

    def get_parent(self, job_str: str) -> Dict:
        """Get job results from the parent job

        :param job_str: This job
        :return: The parent job
        """

    def get_previous(self, job_str: str) -> List[Dict]:
        """Get all job results that can before this job

        :param job_str: This job
        :return: All previous
        """

    def _set_preconditions(self, value: Dict):
        if value is None:
            raise Exception("You may not set an empty value")

        if Store.PIPE_ID not in value:
            raise Exception(f"There must a be {Store.PIPE_ID} field in the dictionary to persist it")

    def parents_like(self, pattern: str) -> List[str]:
        """Get all the parent jobs that match some pattern

        :param pattern: A pattern to match
        :return: Any parents that match this
        """

    def children_like(self, pattern: str) -> List[str]:
        """Get all jobs that match some pattern

        :param pattern: A pattern to match
        :return: Any jobs taht match this
        """

    def is_a_child(self, child: str) -> bool:
        """Is the resource a child of a job

        :param child: The name of the child
        :return: `True` if its a child
        """


class MemoryStore(Store):
    """MemoryStore is a memory-backed store, mainly for testing
    """

    def __init__(self):
        """Use a `Dict` to hold state locally
        """
        super().__init__()
        self.db = {}

    def get(self, job_id: str) -> Dict:
        """This gives back the result of the job store for this entry

        :param job_id: This is a unique ID for this job
        :return: A dictionary containing the output user data
        """
        return self.db[job_id]

    def _set(self, value: Dict) -> None:
        """This updates the job store for this entry

        This Dict must include a `label` key
        :param value: An object to store for this job
        """

        self.db[value[Store.PIPE_ID]] = value

    def exists(self, job_id: str) -> bool:
        """Check if there is a job in the database with this id

        :param job_id:
        :return:
        """
        return job_id in self.db

    def remove(self, job_id: str) -> bool:
        """Delete a job from the database

        :param job_id:
        :return: If the job could be removed
        """
        if job_id not in self.db:
            return False

        del self.db[job_id]
        return True

    def get_parent(self, job_str: str) -> Dict:
        """Get job results from the parent job

        :param job_str: This job
        :return: The parent job
        """
        return self.get(self.get(job_str)[Store.PARENT])

    def get_previous(self, job_str: str) -> List[Dict]:
        """Get all job results that came before this job

        :param job_str: This job
        :return: All previous
        """
        parent = self.get_parent(job_str)
        return [self.get(prev) for prev in parent[Store.EXECUTED]]

    def _match_like(self, pattern: str) -> List[str]:
        matches = [key for key in self.db.keys() if re.match(pattern, key)]
        return matches

    def parents_like(self, pattern: str) -> List[str]:
        """Get all the parent jobs that match some pattern

        :param pattern: A pattern to match
        :return: Any parents that match this
        """
        matches = self._match_like(pattern)
        parents = [match for match in matches if 'j--' not in match]
        return parents

    def children_like(self, pattern: str) -> List[str]:
        """Get all jobs that match some pattern

        :param pattern: A pattern to match
        :return: Any jobs taht match this
        """
        matches = self._match_like(pattern)
        children = [match for match in matches if 'j--' in match]
        return children

    def is_a_child(self, child: str) -> bool:
        """Is the resource a child of a job

        This is a hack, but we dont really care since this
        isnt used for production, just testing.

        :param child: The name of the child
        :return: `True` if its a child
        """
        return 'j--' in child


class Cache:
    """Abstract basecase for a key-value cache."""

    def __setitem__(self, key: str, value: Any) -> None:
        """Add to the cache, if the key exists overwrite it.

        :param key: The key to add to the cache.
        :param value: The value that goes with the key
        """

    def __getitem__(self, key: str) -> Any:
        """Get data from the cache. Returns None if key is missing.

        :param key: The key to look for.
        :returns: The value associated with key.
        """

    def __delitem__(self, key: str) -> None:
        """Delete a key-value pair from the cache.

        :param key: The key to remove.
        """

    def keys(self) -> List[str]:
        """Get the keys in the cache."""


class MemoryCache(Cache):
    """An in-memory key-value cache implemented as a dict."""

    def __init__(self):
        super().__init__()
        self.db = {}

    def __setitem__(self, key: str, value: Any) -> None:
        """Add to the cache, if the key exists overwrite it.

        :param key: The key to add to the cache.
        :param value: The value that goes with the key
        """
        self.db[key] = value

    def __getitem__(self, key: str) -> Any:
        """Get data from the cache. Returns None if key is missing.

        :param key: The key to look for.
        :returns: The value associated with key.
        """
        return self.db.get(key)

    def __delitem__(self, key: str) -> None:
        """Delete a key-value pair from the cache.

        :param key: The key to remove.
        """
        self.db.pop(key, None)

    def keys(self) -> List[str]:
        """Get the keys in the cache."""
        return self.db.keys()


ODIN_STORE_BACKENDS = {}


@optional_params
def register_store_backend(cls, name):
    """Register a function as a plug-in"""
    return register(cls, ODIN_STORE_BACKENDS, name, 'backend')


def create_store_backend(**kwargs):
    name = kwargs.get('backend', 'mongo')
    if name not in ODIN_STORE_BACKENDS:
        LOGGER.info("Loading %s backend", name)
        import_user_module(f'odin.{name}')
    creator_fn = ODIN_STORE_BACKENDS[name]
    return creator_fn(**kwargs)


ODIN_CACHE_BACKENDS = {}


@optional_params
def register_cache_backend(cls, name):
    """Register a function as a plug-in"""
    return register(cls, ODIN_CACHE_BACKENDS, name, 'cache')


def create_cache_backend(**kwargs):
    name = kwargs.get('backend', 'mongo')
    if name not in ODIN_STORE_BACKENDS:
        LOGGER.info("Loading %s backend", name)
        import_user_module(f'odin.{name}')
    creator_fn = ODIN_CACHE_BACKENDS[name]
    return creator_fn(**kwargs)


def write_outputs(output: Dict, label: Optional[str] = None, cred_file: Optional[str] = None) -> None:
    """Write the contents of `output` to the store as your output.

    :param output: The data you want to write, must be savable in mongo.
    :param label: The label used to look up the jobs db entry. Backs off to value stored in $ODIN_TASK_ID
    :param cred_file: The location where the odin cred file lives. Backs off to value stored in $ODIN_CRED
    """
    label = label if label is not None else os.getenv(ODIN_TASK_ENV)
    cred_file = cred_file if cred_file is not None else os.getenv(ODIN_CRED_ENV)
    cred = read_config_file(cred_file)
    store = create_store_backend(**cred['jobs_db'])
    entry = store.get(label)
    out = entry['outputs']
    if out is None:
        entry['outputs'] = output
    else:
        out.update(output)
        entry['outputs'] = output
    store.set(entry)



"""Websocket client to generate a config."""
import argparse
import asyncio
import getpass
import json
import websockets

from baseline.utils import str2bool
from odin import LOGGER, APIField, APIStatus
from odin.client import ODIN_URL, ODIN_PORT
from odin.generate import preprocess_arguments, ALWAYS, IF_NOT_PRESENT, NEVER


async def request_generate_config(ws, config):
    """Use async to open a connection to serve.py and generate a config."""
    async with websockets.connect(ws) as websocket:
        await websocket.send(json.dumps({APIField.COMMAND: 'GENERATE', APIField.REQUEST: config}))

        result = json.loads(await websocket.recv())
        if result[APIField.STATUS] == APIStatus.ERROR:
            LOGGER.error(result)
            return
        if result[APIField.STATUS] == APIStatus.OK:
            LOGGER.info('Generated pipeline is called %s', result[APIField.RESPONSE])


def main():
    """Use `asyncio` to connect to a websocket and request the creation of a pipeline config."""
    parser = argparse.ArgumentParser()
    parser.add_argument("configs", nargs="+", help="The path to the mead config we want to turn into a pipeline.")
    parser.add_argument("--host", default=ODIN_URL, help="The hostname of the odin server.")
    parser.add_argument("--port", default=ODIN_PORT, help="The port of the odin server.")
    parser.add_argument(
        '--scheme',
        choices={'wss', 'ws'},
        default='wss',
        help='Websocket connection protocol, use `wss` for remote connections and `ws` for localhost',
    )
    parser.add_argument("--user", default=getpass.getuser(), help="The prefix of your pipeline name.")
    parser.add_argument(
        "--pipeline-name", "--pipeline_name", help="The name of your pipeline. Defaults to the name of the mead config."
    )
    parser.add_argument("--datasets", help="The location of your custom datasets index file.")
    parser.add_argument("--embeddings", help="The location of you custom embeddings file.")
    parser.add_argument("--train-file", "--train_file", help="The location of a custom train file")
    parser.add_argument("--valid-file", "--valid_file", help="The location of a custom valid file")
    parser.add_argument("--test-file", "--test_file", help="The location of a custom test file")
    parser.add_argument("--task", help="The baseline task for this model")
    parser.add_argument(
        "--models",
        type=int,
        default=1,
        help="The number of times to train the model (or the number of models to train when using hpctl).",
    )
    parser.add_argument("--gpus", type=int, help="The number of GPUs to give to each model training task.")
    parser.add_argument("--metric", default="acc", help="The name of the metric used to compare the results of models.")
    # Including --export-type for back compat
    parser.add_argument(
        "--export-policy",
        "--export_policy",
        "--export-type",
        default=None,
        help="The type of decision to use when deciding if a model should be exported.",
    )
    parser.add_argument(
        "--slack",
        type=str2bool,
        default=True,
        help="Should we send a slack message when the pipeline has finished running?",
    )
    parser.add_argument(
        "--slack-web-hook", "--slack_web_hook", default=None, help="The endpoint for slack messages to go to"
    )
    # parser.add_argument("--git-commit", "--git_commit", type=str2bool, default=False,
    #                     help="Should we commit the exported model with git?")
    parser.add_argument("--mead-image", "--mead_image", help="The name of the image to use for mead training.")
    parser.add_argument(
        "--odin-image", "--odin_image", help="The name of the image to use for odin exporting and odin chores."
    )
    parser.add_argument("--claim-name", "--claim_name", help="The name of the k8s pvc claim.")
    parser.add_argument(
        "--pull-policy",
        "--pull_policy",
        help="The pull policy to use for containers",
        choices={IF_NOT_PRESENT, ALWAYS, NEVER},
        default=ALWAYS,
    )
    parser.add_argument("--clobber", type=str2bool, default=True, help="Should we overwrite a previous pipeline?")
    parser.add_argument("--output-dir", "--output_dir", help="The base dir for where a model should be exported.")
    parser.add_argument("--project", help="The name of the project this model is for.")
    parser.add_argument("--name", help="The name of the model, i.e. intent, sf, etc.")
    parser.add_argument('--modules', help='modules to load', default=[], nargs='+', required=False)
    parser.add_argument('--mead-eval-dataset', "--mead_eval_dataset", help="The dataset to use for mead-eval")
    parser.add_argument('--template', help='An odin-template sample file.')
    parser.add_argument('--hpctl', help='Should hpctl be used in the pipeline to sample configs', action="store_true")
    parser.add_argument('--seed', help='A seed for controlling hpctl', type=int)
    args = parser.parse_args()

    config = preprocess_arguments(args)

    ws = f'{args.scheme}://{args.host}:{args.port}'
    asyncio.get_event_loop().run_until_complete(request_generate_config(ws, config))


if __name__ == "__main__":
    main()



"""Core components for odin
"""
import os
import re
import inspect
import logging
from collections import defaultdict
from string import Template
from typing import Dict, Union, List, Tuple, Any, Optional
import yaml
import shortid
from baseline.utils import listify, is_sequence
from odin.dag import Graph


SHORT_ID = shortid.ShortId()
DEPENDENCY_KEY = 'depends'
TASK_LIST = 'tasks'
ROOT_PATH = 'ROOT_PATH'
WORK_PATH = 'WORK_PATH'
DATA_PATH = 'DATA_PATH'
RUN_PATH = 'RUN_PATH'
TASK_PATH = 'TASK_PATH'
PIPE_ID = 'PIPE_ID'
TASK_IDS = 'TASK_IDS'
TASK_ID = 'TASK_ID'
TASK_NAME = 'TASK_NAME'
K8S_NAME = re.compile(r"[a-z0-9-\.]+")
REFERENCE_KEY = re.compile(r"^\^")


LOGGER = logging.getLogger('odin')


def is_reference(key: Any) -> bool:
    """Is this key a reference to output from elsewhere in the chores.

    A reference has the form ^chore(.output)? where `chore` is the name of the
    node you are consuming output from and `output` is either a name or a
    number to tell which part of the output to take.

    :param key: A key to test
    :returns: `True` if its a reference
    """
    return isinstance(key, str) and key.startswith('^')


def parse_reference(ref: str) -> Tuple[str, ...]:
    """Convert the references in to a list of lookups.

    The reference is assumed to be of the form `^X.Y...` or just `X`

    :param ref: A reference to parse
    :returns: A tuple based on dot splitting and removing the `^`
    """
    parts = ref.split('.')
    parts[0] = REFERENCE_KEY.sub("", parts[0])  # Remove beginning `^` if there.
    return parts


def extract_outputs(path: List[str], results: Dict) -> Union[Any, List[Any]]:
    """Pull data out of results according to ref.

    :param path: The data location.
    :param results: The data to pull content from.
    :returns: The data or None if it could not be found.
    """
    miss = False
    for key in path:
        if key not in results:
            miss = True
            break
        results = results[key]
    return results if not miss else None


def wire_inputs(inputs: Dict, results: Dict, chore: 'Chore') -> Dict:
    """Replace the reference inputs with the output files.

    References are assumed to be of the form `^X.Y` or just `X`.
    If the former, the chore and sub-field of the results Dict
    are returned.  In the latter case, the whole Dict associated
    with the results is returned

    :param inputs: A dictionary of inputs
    :param results: A dictionary of upstream results
    :param chore: The chore function that will be called.
    :returns: The substituted input dictionary
    """
    for key, values in inputs.items():
        if is_sequence(values):
            new_vs = []
            for value in values:
                if is_reference(value):
                    new_vs.append(extract_outputs(parse_reference(value), results))
                else:
                    new_vs.append(value)
            inputs[key] = new_vs
        else:
            if is_reference(values):
                inputs[key] = extract_outputs(parse_reference(values), results)
    # Get the signature of the function
    sig = inspect.signature(chore)
    # Bind the args we populated with inputs
    bound = sig.bind_partial(**inputs)
    for param in sig.parameters.values():
        # Look at all params and if they haven't been bound (they are not
        # present in the bound args and therefore were not in inputs) default
        # them to `None` in inputs. If this param has a default value we
        # don't need to add it to inputs
        if param.name not in bound.arguments and param.default is param.empty:
            inputs[param.name] = None
    return inputs


def create_graph(  # pylint: disable=too-many-nested-blocks,too-many-branches
    task_list: List[Dict], external_inputs: Dict = {}
) -> Graph:
    """Convert task list into a graph.

    There is a link between a chores if a chore references another in
    any of it's inputs or if the chore is listed in a key called `depends`
    (`depends` is used for forced control flow when not a explicit input dep)

    :param task_list: A list of dictionaries describing tasks
    :param external_inputs: A dictionary of external inputs
    :raises ValueError: If a task name contains a `.` or there is a phantom dependency
    :returns: A DAG
    """
    graph: Graph = defaultdict(set)
    name2idx = {task.get('_name', task.get('name')): i for i, task in enumerate(task_list)}
    idx2name = {i: k for k, i in name2idx.items()}
    for name in name2idx:
        if '.' in name:
            raise ValueError(f"Names cannot contain `.` found {name} ")

    for dst, task in enumerate(task_list):
        if DEPENDENCY_KEY in task:
            for src in listify(task[DEPENDENCY_KEY]):
                src = src[1:] if src.startswith('^') else src
                if src not in name2idx:
                    raise ValueError(f"Dependency `{src}` of node `{idx2name[dst]}` not found in graph.")
                graph[name2idx[src]].add(dst)
        for values in task.values():
            values = listify(values)
            for value in values:
                if is_reference(value):
                    lookups = parse_reference(value)
                    src = lookups[0]
                    if src not in external_inputs:
                        if src not in name2idx:
                            raise ValueError(f"Dependency `{src}` of node `{idx2name[dst]}` not found in graph.")
                        graph[name2idx[src]].add(dst)
                    else:
                        LOGGER.info("No dependency required in this graph from %s to %s", dst, src)
    for dst in range(len(task_list)):
        if dst not in graph:
            graph[dst] = set()
    return graph


def format_output(output) -> Union[Dict, List]:
    """Convert the outputs into a consistent format.

    The outputs are dicts. When functions return lists/scalars they are
    converted into dicts with numbers (as str's) for the keys.

    :param output Output to convert
    :return the formatted output
    """
    if is_sequence(output):
        result = {}
        for i, out in enumerate(listify(output)):
            result[str(i)] = out
        output = result
    return output


def _to_kwargs(params: Dict) -> str:
    """Stringify the params as an arg list

    :param params: The params
    :return: A string representation of the parameters in the for `param=arg`.
    """
    return ", ".join(f"{k}={v}" for k, v in params.items() if k != 'name')


def _generate_name(prefix: str) -> str:
    """This generates a new name from the provided prefix suffixed by a shortid

    :param prefix: A provided prefix
    :returns: A unique name that is a combination of the prefix and a shortid
    """
    short_id = SHORT_ID.generate().lower().replace('_', '-')
    return f'{prefix}-{short_id}j'


def validate_pipeline_name(name: str) -> bool:
    """Check if a pipeline name is valid.

    From the k8s docs (https://kubernetes.io/docs/concepts/overview/working-with-objects/names/#names):
        By convention, the names of Kubernetes resources should be up to maximum length of 253 characters and consist
        of lower case alphanumeric characters, -, and ., but certain resources have more specific restrictions.

    :param name: The name to validate
    :returns: Is the name valid of not?
    """
    return K8S_NAME.fullmatch(name) is not None


def _get_child_name(parent_name: str, name: str) -> str:
    """Child pod names are canonical keyed off the unique id of the parent

    :param parent_name: The parent name
    :param name: The child task name
    :returns: A new id that is unique for the child
    """
    return f'{parent_name}--{name}'


def read_pipeline_config(  # pylint: disable=too-many-locals
    work_dir: str, root_dir: str, data_dir: Optional[str] = None, main_file: Optional[str] = None
) -> Tuple[Dict, List[Dict]]:
    """Read in the pipeline configuration from a directory

    The actual flow is specified in a `main.yml` inside this directory.
    The names that are given in the file are relative to the flow, whose actual
    name is some randomized version based on the `name` given.   We need some
    way to refer to these jobs by other jobs in the same flow.  To facilitate,
    use a context and properly substitute user variables which may be present
    in the command arguments

    :param work_dir: A directory pointing to this workflow
    :param root_dir: A directory pointing to all global config info
    :param data_dir: A directory pointing to where this workflow saves data
    :param main_file: Optional config path.  Defaults to `{work_dir}/main.yml`
    :raises ValueError: If there are multiple tasks of the same name.
    :return: A tuple of the context and the interpolated task list
    """
    template_file = main_file if main_file else f'{work_dir}/main.yml'
    if os.path.isfile(template_file):
        LOGGER.info("Loading file: %s", template_file)
        with open(template_file) as read_file:
            flow = yaml.load(read_file, Loader=yaml.FullLoader)
    else:
        LOGGER.info("Loading YAML string: ...%s", format(template_file[-20:]))
        flow = yaml.load(template_file, Loader=yaml.FullLoader)
    basename = flow.get('name', 'flow')
    if not validate_pipeline_name(basename):
        raise ValueError(f"Pipeline name must match {K8S_NAME.pattern}, got {basename}")
    my_id = _generate_name(basename)
    tasks = flow['tasks']
    child_job_ids = [_get_child_name(my_id, task['name']) for task in tasks]
    data_dir = data_dir if data_dir is not None else work_dir
    run_dir = os.path.join(data_dir, my_id)

    context = {
        WORK_PATH: work_dir,
        ROOT_PATH: root_dir,
        PIPE_ID: my_id,
        TASK_IDS: child_job_ids,
        DATA_PATH: data_dir,
        RUN_PATH: run_dir,
    }
    task_names = set()
    for i in range(len(tasks)):
        task_name = tasks[i]['name']
        if task_name in task_names:
            raise ValueError(f"Task names must be unique. Found {task_name} twice.")
        task_names.add(task_name)
        tasks[i]['name'] = child_job_ids[i]
        context[TASK_ID] = child_job_ids[i]
        context[TASK_NAME] = task_name
        task_dir = os.path.join(run_dir, task_name)
        os.makedirs(task_dir, exist_ok=True)
        context[TASK_PATH] = task_dir
        tasks[i]['_name'] = task_name
        for j, arg in enumerate(tasks[i].get('args', [])):
            tasks[i]['args'][j] = Template(arg).substitute(context)
            if arg != tasks[i]['args'][j]:
                LOGGER.info("Interpolated: %s", tasks[i]['args'][j])

    del context[TASK_ID]
    del context[TASK_NAME]
    del context[TASK_PATH]
    return context, tasks
